"""
Copyright 2025 Google LLC
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
     https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

r"""Convert weights from a DeepSeek style model to a MaxText one.

Example cmd:

python3 MaxText/convert_deepseek_ckpt.py --base_model_path <path/to/meta/ckpt> \
    --maxtext_model_path <GCS/path/to/save/new/maxtext/ckpt> --model_size deepseek2-16b
"""

# pylint: disable=line-too-long
# pylint: disable=unsupported-assignment-operation
# pytype: disable=unsupported-operands

import argparse
import pathlib
import os
import gc
import logging

import numpy as np
import torch
import psutil
from tqdm import tqdm

import max_logging
from safetensors import safe_open
import llama_or_mistral_ckpt


MODEL_PARAMS_DICT = {
    "deepseek2-16b": {
        "num_layers": 27,
        "first_num_dense_layers": 1,
        "base_num_query_heads": 16,
        "base_emb_dim": 2048,
        "num_experts": 64,
        "q_lora_rank": 0,
        "kv_lora_rank": 512,
        "qk_nope_head_dim": 128,
        "qk_rope_head_dim": 64,
        "v_head_dim": 128,
    },
    "deepseek3-671b": {
        "num_layers": 61,
        "first_num_dense_layers": 3,
        "base_num_query_heads": 128,
        "base_emb_dim": 7168,
        "num_experts": 256,
        "q_lora_rank": 1536,
        "kv_lora_rank": 512,
        "qk_nope_head_dim": 128,
        "qk_rope_head_dim": 64,
        "v_head_dim": 128,
    },
}


# skip to convert ckpts related to MTP for now
MTP_KEYS_SUFFIX = [
    "enorm.weight",
    "hnorm.weight",
    "eh_proj.weight",
    "shared_head.norm.weight",
    "shared_head.head.weight",
    "model.layers.61.embed_tokens.weight",
]


def _is_key_ending_allowed(key, banned_endings) -> bool:
  """
  Checks if a key's ending is NOT in a list of banned endings.

  Args:
      text: The key to check.
      banned_endings: A list of strings representing banned endings.

  Returns:
      True if the string's ending is not in the banned list, False otherwise.
  """
  for ending in banned_endings:
    if key.endswith(ending):
      return False
  return True


def _hf_to_maxtext_mapping(layer_idx, num_experts, first_num_dense_layers) -> dict:
  """
  Generates a mapping between Hugging Face (HF) and MaxText model weight names.
  HF MLP is using self.down_proj(self.act_fn(self.gate_proj(x)) * self.up_proj(x)).

  Args:
      layer_idx: The index of the current layer.
      num_experts: The number of experts in MoE layers.
      first_num_dense_layers: The number of initial dense layers.

  Returns:
      A dictionary mapping HF weight names to MaxText weight names.
  """
  mapping = {
      "model.embed_tokens.weight": "token_embedder.embedding",
      "lm_head.weight": "logits_dense.kernel",
      "model.norm.weight": "decoder_norm.scale",
  }

  if layer_idx < first_num_dense_layers:
    # Dense layers mapping
    mapping.update(
        {
            f"model.layers.{layer_idx}.input_layernorm.weight": f"dense_layers.{layer_idx}.pre_self_attention_layer_norm.scale",
            f"model.layers.{layer_idx}.post_attention_layernorm.weight": f"dense_layers.{layer_idx}.post_self_attention_layer_norm.scale",
            f"model.layers.{layer_idx}.self_attn.q_proj.weight": f"dense_layers.{layer_idx}.self_attention.query.kernel",
            f"model.layers.{layer_idx}.self_attn.q_a_proj.weight": f"dense_layers.{layer_idx}.self_attention.wq_a.kernel",
            f"model.layers.{layer_idx}.self_attn.q_a_layernorm.weight": f"dense_layers.{layer_idx}.self_attention.q_norm.scale",
            f"model.layers.{layer_idx}.self_attn.q_b_proj.weight": f"dense_layers.{layer_idx}.self_attention.wq_b.kernel",
            f"model.layers.{layer_idx}.self_attn.kv_a_proj_with_mqa.weight": f"dense_layers.{layer_idx}.self_attention.wkv_a.kernel",
            f"model.layers.{layer_idx}.self_attn.kv_b_proj.weight": f"dense_layers.{layer_idx}.self_attention.wkv_b.kernel",
            f"model.layers.{layer_idx}.self_attn.o_proj.weight": f"dense_layers.{layer_idx}.self_attention.out.kernel",
            f"model.layers.{layer_idx}.mlp.gate_proj.weight": f"dense_layers.{layer_idx}.mlp.wi_0.kernel",
            f"model.layers.{layer_idx}.mlp.up_proj.weight": f"dense_layers.{layer_idx}.mlp.wi_1.kernel",
            f"model.layers.{layer_idx}.mlp.down_proj.weight": f"dense_layers.{layer_idx}.mlp.wo.kernel",
            f"model.layers.{layer_idx}.self_attn.kv_a_layernorm.weight": f"dense_layers.{layer_idx}.self_attention.kv_norm.scale",
        }
    )
  else:
    # MoE layers mapping
    moe_layer_idx_in_maxtext = layer_idx - first_num_dense_layers
    for expert_idx in range(num_experts):
      mapping.update(
          {
              f"model.layers.{layer_idx}.mlp.experts.{expert_idx}.down_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.MoeBlock_0.{expert_idx}.wo",
              f"model.layers.{layer_idx}.mlp.experts.{expert_idx}.gate_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.MoeBlock_0.{expert_idx}.wi_0",
              f"model.layers.{layer_idx}.mlp.experts.{expert_idx}.up_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.MoeBlock_0.{expert_idx}.wi_1",
          }
      )
    mapping.update(
        {
            f"model.layers.{layer_idx}.mlp.gate.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.MoeBlock_0.gate.kernel",
            f"model.layers.{layer_idx}.mlp.gate.e_score_correction_bias": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.MoeBlock_0.gate.bias",
            f"model.layers.{layer_idx}.mlp.shared_experts.down_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.shared_experts.wo.kernel",
            f"model.layers.{layer_idx}.mlp.shared_experts.gate_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.shared_experts.wi_0.kernel",
            f"model.layers.{layer_idx}.mlp.shared_experts.up_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.DeepSeekMoeBlock_0.shared_experts.wi_1.kernel",
            f"model.layers.{layer_idx}.self_attn.q_a_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.wq_a.kernel",
            f"model.layers.{layer_idx}.self_attn.q_a_layernorm.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.q_norm.scale",
            f"model.layers.{layer_idx}.self_attn.q_b_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.wq_b.kernel",
            f"model.layers.{layer_idx}.self_attn.kv_a_layernorm.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.kv_norm.scale",
            f"model.layers.{layer_idx}.self_attn.kv_a_proj_with_mqa.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.wkv_a.kernel",
            f"model.layers.{layer_idx}.self_attn.kv_b_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.wkv_b.kernel",
            f"model.layers.{layer_idx}.self_attn.o_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.out.kernel",
            f"model.layers.{layer_idx}.self_attn.q_proj.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.self_attention.query.kernel",
            f"model.layers.{layer_idx}.input_layernorm.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.pre_self_attention_layer_norm.scale",
            f"model.layers.{layer_idx}.post_attention_layernorm.weight": f"moe_layers.{moe_layer_idx_in_maxtext}.post_self_attention_layer_norm.scale",
        }
    )
  return mapping


def _convert_huggingface_to_jax_weights(base_model_path, model_params, mem_info) -> dict:
  """Convert Huggingface Checkpoint to Jax."""
  base_num_decoder_layers = model_params["num_layers"]
  first_num_dense_layers = model_params["first_num_dense_layers"]
  base_num_query_heads = model_params["base_num_query_heads"]
  base_emb_dim = model_params["base_emb_dim"]
  num_experts = model_params["num_experts"]
  q_lora_rank = model_params["q_lora_rank"]
  kv_lora_rank = model_params["kv_lora_rank"]
  qk_nope_head_dim = model_params["qk_nope_head_dim"]
  qk_rope_head_dim = model_params["qk_rope_head_dim"]
  v_head_dim = model_params["v_head_dim"]

  ckpt_paths = sorted(pathlib.Path(base_model_path).glob("[!.]*.safetensors"))
  chkpt_vars = {}
  for i, ckpt_path in enumerate(ckpt_paths):
    max_logging.log(f"Loading checkpoint {i+1} of {len(ckpt_paths)} ...")
    with safe_open(ckpt_path, framework="pt", device="cpu") as f:
      for key in f.keys():
        parts = key.split(".")
        layer = int(parts[2]) if "layers" in key else 0
        if key.endswith("_scale_inv"):
          raise ValueError("fp8 checkpoint is not supported.")
        if _is_key_ending_allowed(key, MTP_KEYS_SUFFIX):
          mapped_key = _hf_to_maxtext_mapping(layer, num_experts, first_num_dense_layers)[key]
          chkpt_vars[mapped_key] = f.get_tensor(key)

  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  # initialize the data structure for storing jax_weights
  jax_weights = {
      "decoder": {
          "dense_layers": {
              "mlp": {
                  "wi_0": {"kernel": None},
                  "wi_1": {"kernel": None},
                  "wo": {"kernel": None},
              },
              "self_attention": {
                  "kv_norm": {"scale": None},
                  "wkv_a": {"kernel": None},
                  "wkv_b": {"kernel": None},
                  "out": {"kernel": None},
              },
              "pre_self_attention_layer_norm": {"scale": None},
              "post_self_attention_layer_norm": {"scale": None},
          },
          "moe_layers": {
              "DeepSeekMoeBlock_0": {
                  "MoeBlock_0": {
                      "wi_0": None,
                      "wi_1": None,
                      "wo": None,
                      "gate": {"kernel": None},
                  },
                  "shared_experts": {
                      "wi_0": {"kernel": None},
                      "wi_1": {"kernel": None},
                      "wo": {"kernel": None},
                  },
              },
              "self_attention": {
                  "kv_norm": {"scale": None},
                  "wkv_a": {"kernel": None},
                  "wkv_b": {"kernel": None},
                  "out": {"kernel": None},
              },
              "pre_self_attention_layer_norm": {"scale": None},
              "post_self_attention_layer_norm": {"scale": None},
          },
          "decoder_norm": {"scale": None},
          "logits_dense": {"kernel": None},
      },
      "token_embedder": {"embedding": None},
  }

  # decoder norm scale ###########################################
  max_logging.log("Processing decoder norm scale")
  jax_weights["decoder"]["decoder_norm"]["scale"] = chkpt_vars["decoder_norm.scale"].to(torch.float16).numpy()
  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  # logits dense #################################################
  max_logging.log("Processing logits dense")
  jax_weights["decoder"]["logits_dense"]["kernel"] = chkpt_vars["logits_dense.kernel"].to(torch.float16).numpy().transpose()
  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  # token embedding ##############################################
  max_logging.log("Processing token embeddings")
  jax_weights["token_embedder"]["embedding"] = chkpt_vars["token_embedder.embedding"].to(torch.float16).numpy()
  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  layers = {
      "dense_layers": first_num_dense_layers,
      "moe_layers": base_num_decoder_layers - first_num_dense_layers,
  }
  # self attention and normalization ###############################################
  max_logging.log("Processing self attention and normalization in dense layer")
  for layer_key, layer_value in layers.items():
    self_attention = jax_weights["decoder"][f"{layer_key}"]["self_attention"]
    pre_self_attention_layer_norm = jax_weights["decoder"][f"{layer_key}"]["pre_self_attention_layer_norm"]
    post_self_attention_layer_norm = jax_weights["decoder"][f"{layer_key}"]["post_self_attention_layer_norm"]

    for layer_idx in tqdm(range(layer_value), desc=layer_key, leave=False):
      pre_self_attention = (
          chkpt_vars[f"{layer_key}.{layer_idx}.pre_self_attention_layer_norm.scale"].to(torch.float16).numpy()
      )
      post_self_attention = (
          chkpt_vars[f"{layer_key}.{layer_idx}.post_self_attention_layer_norm.scale"].to(torch.float16).numpy()
      )
      kv_norm = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.kv_norm.scale"].to(torch.float16).numpy().transpose()
      wkv_a = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.wkv_a.kernel"].to(torch.float16).numpy().transpose()
      wkv_b = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.wkv_b.kernel"].to(torch.float16).numpy().transpose()
      out = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.out.kernel"].to(torch.float16).numpy().transpose()
      if q_lora_rank != 0:
        q_norm = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.q_norm.scale"].to(torch.float16).numpy()
        wq_a = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.wq_a.kernel"].to(torch.float16).numpy().transpose()
        wq_b = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.wq_b.kernel"].to(torch.float16).numpy().transpose()
      else:
        query = chkpt_vars[f"{layer_key}.{layer_idx}.self_attention.query.kernel"].to(torch.float16).numpy().transpose()

      # reshape to match maxtext
      wkv_b = np.reshape(wkv_b, [kv_lora_rank, base_num_query_heads, (qk_nope_head_dim + v_head_dim)])
      out = np.reshape(out, [base_num_query_heads, v_head_dim, base_emb_dim])
      if q_lora_rank != 0:
        wq_b = np.reshape(wq_b, [q_lora_rank, base_num_query_heads, (qk_nope_head_dim + qk_rope_head_dim)])
      else:
        query = np.reshape(query, [base_emb_dim, base_num_query_heads, (qk_nope_head_dim + qk_rope_head_dim)])

      # initialization
      if self_attention["kv_norm"]["scale"] is None:
        stack_shape = (layer_value,)
        self_attention["kv_norm"]["scale"] = np.zeros(stack_shape + kv_norm.shape, dtype=np.float16)
        self_attention["wkv_a"]["kernel"] = np.zeros(stack_shape + wkv_a.shape, dtype=np.float16)
        self_attention["wkv_b"]["kernel"] = np.zeros(stack_shape + wkv_b.shape, dtype=np.float16)
        self_attention["out"]["kernel"] = np.zeros(stack_shape + out.shape, dtype=np.float16)
        pre_self_attention_layer_norm["scale"] = np.zeros(stack_shape + pre_self_attention.shape, dtype=np.float16)
        post_self_attention_layer_norm["scale"] = np.zeros(stack_shape + post_self_attention.shape, dtype=np.float16)
        if q_lora_rank != 0:
          self_attention.update(
              {
                  "q_norm": {"scale": None},
                  "wq_a": {"kernel": None},
                  "wq_b": {"kernel": None},
              }
          )
          self_attention["q_norm"]["scale"] = np.zeros(stack_shape + q_norm.shape, dtype=np.float16)
          self_attention["wq_a"]["kernel"] = np.zeros(stack_shape + wq_a.shape, dtype=np.float16)
          self_attention["wq_b"]["kernel"] = np.zeros(stack_shape + wq_b.shape, dtype=np.float16)
        else:
          self_attention.update({"query": {"kernel": None}})
          self_attention["query"]["kernel"] = np.zeros(stack_shape + query.shape, dtype=np.float16)

      self_attention["kv_norm"]["scale"][layer_idx, ...] = kv_norm
      self_attention["wkv_a"]["kernel"][layer_idx, ...] = wkv_a
      self_attention["wkv_b"]["kernel"][layer_idx, ...] = wkv_b
      self_attention["out"]["kernel"][layer_idx, ...] = out
      pre_self_attention_layer_norm["scale"][layer_idx, ...] = pre_self_attention
      post_self_attention_layer_norm["scale"][layer_idx, ...] = post_self_attention
      if q_lora_rank != 0:
        self_attention["q_norm"]["scale"][layer_idx, ...] = q_norm
        self_attention["wq_a"]["kernel"][layer_idx, ...] = wq_a
        self_attention["wq_b"]["kernel"][layer_idx, ...] = wq_b
      else:
        self_attention["query"]["kernel"][layer_idx, ...] = query

    # re-order to fit maxtext
    self_attention["kv_norm"]["scale"] = np.transpose(self_attention["kv_norm"]["scale"], axes=(1, 0))
    self_attention["wkv_a"]["kernel"] = np.transpose(self_attention["wkv_a"]["kernel"], axes=(1, 0, 2))
    self_attention["wkv_b"]["kernel"] = np.transpose(self_attention["wkv_b"]["kernel"], axes=(1, 0, 2, 3))
    self_attention["out"]["kernel"] = np.transpose(self_attention["out"]["kernel"], axes=(1, 0, 2, 3))
    pre_self_attention_layer_norm["scale"] = np.transpose(pre_self_attention_layer_norm["scale"], axes=(1, 0))
    post_self_attention_layer_norm["scale"] = np.transpose(post_self_attention_layer_norm["scale"], axes=(1, 0))
    if q_lora_rank != 0:
      self_attention["q_norm"]["scale"] = np.transpose(self_attention["q_norm"]["scale"], axes=(1, 0))
      self_attention["wq_a"]["kernel"] = np.transpose(self_attention["wq_a"]["kernel"], axes=(1, 0, 2))
      self_attention["wq_b"]["kernel"] = np.transpose(self_attention["wq_b"]["kernel"], axes=(1, 0, 2, 3))
    else:
      self_attention["query"]["kernel"] = np.transpose(self_attention["query"]["kernel"], axes=(1, 0, 2, 3))

    jax_weights["decoder"][f"{layer_key}"]["self_attention"] = self_attention
    jax_weights["decoder"][f"{layer_key}"]["pre_self_attention_layer_norm"] = pre_self_attention_layer_norm
    jax_weights["decoder"][f"{layer_key}"]["post_self_attention_layer_norm"] = post_self_attention_layer_norm
    logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  # layer weights ################################################
  max_logging.log("Processing layer weights")
  for layer_key, layer_value in layers.items():
    if layer_key == "dense_layers":
      mlp = jax_weights["decoder"][f"{layer_key}"]["mlp"]
      for layer_idx in tqdm(range(layer_value), desc=layer_key, leave=False):
        wi_0 = chkpt_vars[f"{layer_key}.{layer_idx}.mlp.wi_0.kernel"].to(torch.float16).numpy().transpose()
        wi_1 = chkpt_vars[f"{layer_key}.{layer_idx}.mlp.wi_1.kernel"].to(torch.float16).numpy().transpose()
        wo = chkpt_vars[f"{layer_key}.{layer_idx}.mlp.wo.kernel"].to(torch.float16).numpy().transpose()
        # initialization
        if mlp["wo"]["kernel"] is None:
          stack_shape = (layer_value,)
          mlp["wi_0"]["kernel"] = np.zeros(stack_shape + wi_0.shape, dtype=np.float16)
          mlp["wi_1"]["kernel"] = np.zeros(stack_shape + wi_1.shape, dtype=np.float16)
          mlp["wo"]["kernel"] = np.zeros(stack_shape + wo.shape, dtype=np.float16)
        mlp["wi_0"]["kernel"][layer_idx, ...] = wi_0
        mlp["wi_1"]["kernel"][layer_idx, ...] = wi_1
        mlp["wo"]["kernel"][layer_idx, ...] = wo

      mlp["wi_0"]["kernel"] = np.transpose(mlp["wi_0"]["kernel"], axes=(1, 0, 2))
      mlp["wi_1"]["kernel"] = np.transpose(mlp["wi_1"]["kernel"], axes=(1, 0, 2))
      mlp["wo"]["kernel"] = np.transpose(mlp["wo"]["kernel"], axes=(1, 0, 2))
      jax_weights["decoder"][f"{layer_key}"]["mlp"] = mlp
      logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))
    else:
      moe = jax_weights["decoder"][f"{layer_key}"]["DeepSeekMoeBlock_0"]
      for layer_idx in tqdm(range(layer_value), desc=layer_key, leave=False):
        if q_lora_rank != 0:
          gate_bias = (
              chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.MoeBlock_0.gate.bias"]
              .to(torch.float16)
              .numpy()
              .transpose()
          )
        gate = (
            chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.MoeBlock_0.gate.kernel"]
            .to(torch.float16)
            .numpy()
            .transpose()
        )
        shared_wi_0 = (
            chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.shared_experts.wi_0.kernel"]
            .to(torch.float16)
            .numpy()
            .transpose()
        )
        shared_wi_1 = (
            chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.shared_experts.wi_1.kernel"]
            .to(torch.float16)
            .numpy()
            .transpose()
        )
        shared_wo = (
            chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.shared_experts.wo.kernel"]
            .to(torch.float16)
            .numpy()
            .transpose()
        )

        # initialization
        if moe["MoeBlock_0"]["gate"]["kernel"] is None:
          stack_shape = (layer_value,)
          if q_lora_rank != 0:
            moe["MoeBlock_0"]["gate"].update({"bias": None})
            moe["MoeBlock_0"]["gate"]["bias"] = np.zeros(stack_shape + gate_bias.shape, dtype=np.float16)
          moe["MoeBlock_0"]["gate"]["kernel"] = np.zeros(stack_shape + gate.shape, dtype=np.float16)
          moe["shared_experts"]["wi_0"]["kernel"] = np.zeros(stack_shape + shared_wi_0.shape, dtype=np.float16)
          moe["shared_experts"]["wi_1"]["kernel"] = np.zeros(stack_shape + shared_wi_1.shape, dtype=np.float16)
          moe["shared_experts"]["wo"]["kernel"] = np.zeros(stack_shape + shared_wo.shape, dtype=np.float16)

        if q_lora_rank != 0:
          moe["MoeBlock_0"]["gate"]["bias"][layer_idx, ...] = gate_bias
        moe["MoeBlock_0"]["gate"]["kernel"][layer_idx, ...] = gate
        moe["shared_experts"]["wi_0"]["kernel"][layer_idx, ...] = shared_wi_0
        moe["shared_experts"]["wi_1"]["kernel"][layer_idx, ...] = shared_wi_1
        moe["shared_experts"]["wo"]["kernel"][layer_idx, ...] = shared_wo

      # re-order
      if q_lora_rank != 0:
        moe["MoeBlock_0"]["gate"]["bias"] = np.transpose(moe["MoeBlock_0"]["gate"]["bias"], axes=(1, 0))
      moe["MoeBlock_0"]["gate"]["kernel"] = np.transpose(moe["MoeBlock_0"]["gate"]["kernel"], axes=(1, 0, 2))
      moe["shared_experts"]["wi_0"]["kernel"] = np.transpose(moe["shared_experts"]["wi_0"]["kernel"], axes=(1, 0, 2))
      moe["shared_experts"]["wi_1"]["kernel"] = np.transpose(moe["shared_experts"]["wi_1"]["kernel"], axes=(1, 0, 2))
      moe["shared_experts"]["wo"]["kernel"] = np.transpose(moe["shared_experts"]["wo"]["kernel"], axes=(1, 0, 2))

      for layer_idx in tqdm(range(layer_value), desc=layer_key, leave=False):
        for k in tqdm(range(num_experts), desc="experts", leave=False):
          wi_0 = (
              chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.MoeBlock_0.{k}.wi_0"]
              .to(torch.float16)
              .numpy()
              .transpose()
          )
          wi_1 = (
              chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.MoeBlock_0.{k}.wi_1"]
              .to(torch.float16)
              .numpy()
              .transpose()
          )
          wo = (
              chkpt_vars[f"{layer_key}.{layer_idx}.DeepSeekMoeBlock_0.MoeBlock_0.{k}.wo"]
              .to(torch.float16)
              .numpy()
              .transpose()
          )

          if moe["MoeBlock_0"]["wi_0"] is None:
            stack_shape = (num_experts, layer_value)
            moe["MoeBlock_0"]["wi_0"] = np.zeros(stack_shape + wi_0.shape, dtype=np.float16)
            moe["MoeBlock_0"]["wi_1"] = np.zeros(stack_shape + wi_1.shape, dtype=np.float16)
            moe["MoeBlock_0"]["wo"] = np.zeros(stack_shape + wo.shape, dtype=np.float16)
          moe["MoeBlock_0"]["wi_0"][k, layer_idx, ...] = wi_0
          moe["MoeBlock_0"]["wi_1"][k, layer_idx, ...] = wi_1
          moe["MoeBlock_0"]["wo"][k, layer_idx, ...] = wo

      jax_weights["decoder"][f"{layer_key}"]["DeepSeekMoeBlock_0"] = moe
      logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))

  del chkpt_vars
  gc.collect()
  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))
  return jax_weights


def _convert_to_jax_weights(base_model_path, model_size, mem_info) -> dict:
  """
  Function to convert the checkpoint at base_model_path into Orbax checkpoint
  for MaxText and output jax_weights ready for MaxText.

  Args:
      base_model_path: Path to the Hugging Face model checkpoint.
      model_size: Model size key in MODEL_PARAMS_DICT.
      mem_info: A process instance used for memory tracking.

  Returns:
      The converted JAX weights.
  """
  model_params = MODEL_PARAMS_DICT[model_size]
  logging.debug("Memory usage: %f GB", mem_info.memory_info().rss / (1024**3))
  max_logging.log(f"Loading the base model from {base_model_path}")
  return _convert_huggingface_to_jax_weights(base_model_path, model_params, mem_info)


def main() -> None:
  parser = argparse.ArgumentParser(description="Convert and save DeepSeek model weights.")
  parser.add_argument("--base_model_path", type=str, required=True)
  parser.add_argument("--maxtext_model_path", type=str, required=True)
  parser.add_argument("--model_size", type=str, required=True)
  parser.add_argument("--simulated_cpu_devices_count", type=int, required=False, default=16)
  args = parser.parse_args()

  if args.model_size not in MODEL_PARAMS_DICT:
    raise NotImplementedError(f"Model '{args.model_size}' is not supported.")

  os.environ["JAX_PLATFORMS"] = "cpu"
  os.environ["XLA_FLAGS"] = f"--xla_force_host_platform_device_count={args.simulated_cpu_devices_count}"
  mem_info = psutil.Process()
  llama_or_mistral_ckpt.save_weights_to_checkpoint(
      args.maxtext_model_path,
      _convert_to_jax_weights(args.base_model_path, args.model_size, mem_info),
      args.simulated_cpu_devices_count,
  )


if __name__ == "__main__":
  main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model server warmup utils."""

import jax.numpy as jnp
import concurrent.futures
from typing import Any, Optional
import logging
from jetstream.engine import engine_api, token_utils


def layout_params_and_compile_executables(
    prefill_engines: Optional[list[engine_api.JetStreamEngine]] = None,
    generate_engines: Optional[list[engine_api.JetStreamEngine]] = None,
    prefill_params: Optional[list[Any]] = None,
    generate_params: Optional[list[Any]] = None,
) -> bool:
  """Organizes the engines and executables.

  Args:
      prefill_engines: Prefill only engines.
      generate_engines: Generate only engines.
      prefill_params: Prefill only params.
      generate_params: Generate only params.
  """
  prefill_engines = prefill_engines if prefill_engines else []
  generate_engines = generate_engines if generate_engines else []
  prefill_params = prefill_params if prefill_params else []
  generate_params = generate_params if generate_params else []

  any_prefill_engine = None
  any_prefill_params = None

  prefills_compiled = []
  inserts_generate_compiled = []

  for i, pe in enumerate(prefill_engines):
    any_prefill_engine = pe
    any_prefill_params = prefill_params[i]
    prefill_compiled = initialize_prefill_jit_cache(
        prefill_engine=pe,
        prefill_params=prefill_params[i],
        prefill_idx=i,
    )
    prefills_compiled.append(prefill_compiled)

  for i, ge in enumerate(generate_engines):
    insert_generate_compiled = initialize_insert_generate_jit_cache(
        prefill_engine=any_prefill_engine,
        generate_engine=ge,
        prefill_params=any_prefill_params,
        generate_params=generate_params[i],
        generate_idx=i,
    )
    inserts_generate_compiled.append([insert_generate_compiled])

  if all(prefills_compiled) and all(inserts_generate_compiled):
    return True
  return False


def initialize_prefill_jit_cache(
    *,
    prefill_engine: engine_api.JetStreamEngine,
    prefill_params: Any,
    prefill_idx: int,
):
  """Precompile all prefill functions in parallel.
  If we don't do this, then when a new request triggers a new prefill bucket it
  will take a very long time for that query to come back.

  Args:
      prefill_engine: A prefill engine to be compiled for.
      prefill_params: The associated prefill parameters.
      prefill_idx: Which prefill engine it is.
  """
  prefill_buckets = token_utils.DEFAULT_PREFILL_BUCKETS
  prefill_buckets = [
      bucket
      for bucket in prefill_buckets
      if bucket <= prefill_engine.max_prefill_length
  ]
  prefill_engine.prefill_buckets = prefill_buckets
  if prefill_engine.max_prefill_length not in prefill_buckets:
    prefill_buckets.append(prefill_engine.max_prefill_length)

  def compile_prefill(length):
    padded_tokens, true_length = jnp.ones((length), dtype="int32"), length

    _, _ = prefill_engine.prefill(
        params=prefill_params,
        padded_tokens=padded_tokens,
        true_length=true_length,
    )

    _, _ = prefill_engine.prefill_multisampling(
        params=prefill_params,
        padded_tokens=padded_tokens,
        true_length=true_length,
    )

    logging.info(
        "---------Prefill engine %d compiled for prefill length %d.---------",
        prefill_idx,
        length,
    )

  logging.info("---------Prefill compilation %d begun.---------", prefill_idx)

  with concurrent.futures.ThreadPoolExecutor(
      max_workers=len(prefill_buckets)
  ) as executor:
    _ = executor.map(compile_prefill, prefill_buckets)

  prefill_engine.warm = True

  logging.info(
      "---------Prefill compilation %d complete.---------", prefill_idx
  )

  return prefill_engine.warm


def initialize_insert_generate_jit_cache(
    *,
    prefill_engine: engine_api.JetStreamEngine,
    generate_engine: engine_api.JetStreamEngine,
    prefill_params: Any,
    generate_params: Any,
    generate_idx: int,
):
  """Initialiszes jit cache for insert and generate.

  Args:
      generate_engine: A generate engine to be compiled for.
      generate_params: The associated parameters.
      generate_idx: Which generate engine it is.
  """

  prefill_buckets = token_utils.DEFAULT_PREFILL_BUCKETS
  prefill_buckets = [
      bucket
      for bucket in prefill_buckets
      if bucket <= generate_engine.max_prefill_length
  ]
  generate_engine.prefill_buckets = prefill_buckets
  if generate_engine.max_prefill_length not in prefill_buckets:
    prefill_buckets.append(generate_engine.max_prefill_length)

  decode_state = generate_engine.init_decode_state()

  def compile_insert(length):
    padded_tokens, true_length = jnp.ones((length), dtype="int32"), length

    prefill, _ = prefill_engine._downstream_engine.prefill(  # pylint: disable=protected-access
        params=prefill_params,
        padded_tokens=padded_tokens,
        true_length=true_length,
    )

    generate_engine.insert(prefix=prefill, decode_state=decode_state, slot=0)

    generate_engine.bulk_insert(
        prefix=prefill, decode_state=decode_state, slots=[0]
    )

    logging.info(
        "---------Generate engine %d compiled for insert length %d.---------",
        generate_idx,
        length,
    )

  def compile_generate():

    logging.info(
        "---------Generate compilation %d begun.---------", generate_idx
    )

    generate_engine._downstream_engine.generate(  # pylint: disable=protected-access
        params=generate_params,
        decode_state=decode_state,
    )

    logging.info(
        "---------Generate engine %d compiled.---------",
        generate_idx,
    )

    logging.info(
        "---------Generate compilation %d complete.---------", generate_idx
    )

  logging.info(
      "---------Insertion generation compilation %d begun.---------",
      generate_idx,
  )

  compile_generate()

  logging.info(
      "---------Generate engine %d compiled generation step.---------",
      generate_idx,
  )

  with concurrent.futures.ThreadPoolExecutor(
      max_workers=len(prefill_buckets)
  ) as executor:
    _ = executor.map(compile_insert, prefill_buckets)

  generate_engine.warm = True

  logging.info(
      "---------Insertion generation compilation %d complete.---------",
      generate_idx,
  )

  return generate_engine.warm



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests functionality of the token processing utils using mock engine vocab."""

import numpy as np
import unittest
from jetstream.engine import engine_api
from jetstream.engine import mock_utils
from jetstream.engine import token_utils


class UtilsTest(unittest.TestCase):

  def test_speculations_with_multi_sample_slots(self, samples_per_slot=2):
    # [4, 1]
    mock_tokens = np.array(
        [
            [0, ord("A")],
            [ord("A"), ord("D")],
            [ord("T"), ord("3")],
            [ord("A"), 1],
        ]
    ).astype(np.int32)
    mock_valid_tokens = np.ones_like(mock_tokens, dtype=np.int32)
    mock_lengths = np.ones(mock_tokens.shape[0], dtype=np.int32) * 2
    # completion is 'per slot' because we track it for a given request.
    mock_complete = np.zeros(
        (mock_tokens.shape[0] // samples_per_slot), dtype=np.int32
    )
    data = np.concatenate(
        [
            mock_tokens,
            mock_valid_tokens,
            mock_lengths[:, None],
        ],
        axis=-1,
    )
    speculations = mock_tokens.shape[1]
    result_tokens = engine_api.ResultTokens(
        data=data.astype(np.int32),
        tokens_idx=(0, speculations),
        valid_idx=(speculations, 2 * speculations),
        length_idx=(2 * speculations, 2 * speculations + 1),
        samples_per_slot=2,
    )
    vocab = mock_utils.TestVocab()
    per_channel, complete = token_utils.process_result_tokens(
        tokenizer=vocab,
        slots=0,
        slot_max_length=4,
        result_tokens=result_tokens,
        complete=mock_complete,
        is_client_side_tokenization=False,
    )
    np.testing.assert_equal(complete, np.array([1, 0]))

    text_output = [
        mock_utils.TestVocab().decode(row.token_ids) for row in per_channel
    ]
    assert not text_output[0]  # i.e. == '', because of the pad.
    assert text_output[1] == "AD"
    mock_complete = np.zeros(
        (mock_tokens.shape[0] // samples_per_slot), dtype=np.int32
    )
    per_channel, complete = token_utils.process_result_tokens(
        tokenizer=vocab,
        slots=1,
        slot_max_length=4,
        result_tokens=result_tokens,
        complete=mock_complete,
        is_client_side_tokenization=False,
    )
    text_output = [
        mock_utils.TestVocab().decode(row.token_ids) for row in per_channel
    ]
    assert text_output[0] == "T3"
    assert text_output[1] == "A"  # second token is padded.
    np.testing.assert_equal(complete, np.array([0, 1]))

  def test_mock_utils(self):
    vocab = mock_utils.TestVocab()
    # test encode()
    with self.assertRaises(NotImplementedError):
      vocab.encode("AB")
    # test encode_tf()
    token_ids = vocab.encode_tf("AB")
    np.testing.assert_equal(token_ids, np.array([65, 66]))
    # test decode()
    ids = np.array([ord("A")])
    expected = "A"
    result = vocab.decode(ids)
    self.assertEqual(result, expected)
    # test decode_tf()
    ids = np.array([[ord("A")]])
    expected = ["A"]
    result_tf = vocab.decode_tf(ids)
    self.assertEqual(result_tf, expected)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from absl.testing import absltest
import numpy as np
import math
import jax
from jax import numpy as jnp
from inference import nn
from inference.kernel.attention_ops import vanilla_prefill_gqa
from inference.kernel.attention.tpu.chunked_prefill_attention import chunked_prefill_attention


class ChunkedPrefillTest(absltest.TestCase):

  def test(self):
    num_attn_heads = 4
    num_kv_heads = 2
    head_dim = 128
    total_page_num = 16
    page_size = 16

    rng = jax.random.key(0)

    attn_layer = nn.AttentionOps(num_attn_heads, num_kv_heads, head_dim)

    kv_cache = nn.KVCache(
        k=jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
        v=jnp.zeros((num_kv_heads, total_page_num, page_size, head_dim)),
    )

    prefill_len = 6 * page_size
    prefill_non_padding_len = 4 * page_size + 3

    q = jax.random.uniform(rng, (prefill_len, num_attn_heads, head_dim))

    rng_1, rng_2 = jax.random.split(rng)
    k_to_save, v_to_save = (
        jax.random.uniform(rng_1, (prefill_len, num_kv_heads, head_dim)),
        jax.random.uniform(rng_2, (prefill_len, num_kv_heads, head_dim)),
    )

    num_pages_with_padding = math.ceil(prefill_len / page_size)
    page_table = jnp.array(([i for i in range(num_pages_with_padding)]))
    kv_cache = attn_layer._write_prefill_kv_to_kv_cache(
        k_to_save,
        v_to_save,
        kv_cache,
        prefill_non_padding_len,
        page_table,
    )

    chunk_size = 2 * page_size

    expected_output = vanilla_prefill_gqa(q, k_to_save, v_to_save)

    num_active_pages_per_prefill = math.ceil(
        prefill_non_padding_len / page_size
    )
    compute_times = math.ceil(prefill_non_padding_len / chunk_size)

    for i in range(compute_times):
      idx = i * chunk_size
      length = idx + chunk_size
      length = min(length, prefill_non_padding_len)
      chunk_output = chunked_prefill_attention(
          q[idx : idx + chunk_size], kv_cache.k, kv_cache.v, length, page_table
      )

      if i < compute_times - 1:
        np.testing.assert_allclose(
            np.array(chunk_output),
            np.array(expected_output[idx : idx + chunk_size]),
            rtol=4e-03,
            atol=1e-03,
        )
      if i == compute_times - 1:
        offset = prefill_non_padding_len % chunk_size
        np.testing.assert_allclose(
            np.array(chunk_output[:offset]),
            np.array(expected_output[idx : idx + offset]),
            rtol=4e-03,
            atol=1e-03,
        )

    num_pages_per_prefill = prefill_len // page_size
    for i in range(num_active_pages_per_prefill):
      idx = i * page_size
      np.testing.assert_equal(
          np.array(kv_cache.k[:, page_table[i], :, :][:, :, :]),
          np.array(k_to_save[idx : idx + page_size, :, :].transpose(1, 0, 2)),
      )
      np.testing.assert_equal(
          np.array(kv_cache.v[:, page_table[i], :, :][:, :, :]),
          np.array(v_to_save[idx : idx + page_size, :, :].transpose(1, 0, 2)),
      )

    for i in range(num_active_pages_per_prefill, num_pages_per_prefill):
      idx = i * page_size
      np.testing.assert_equal(
          np.array(kv_cache.v[:, page_table[i], :, :][:, None, :, :]),
          np.array(jnp.zeros((num_kv_heads, 1, page_size, head_dim))),
      )

      np.testing.assert_equal(
          np.array(kv_cache.v[:, page_table[i], :, :][:, None, :, :]),
          np.array(jnp.zeros((num_kv_heads, 1, page_size, head_dim))),
      )


if __name__ == "__main__":
  absltest.main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from jax.tree_util import register_pytree_node


def register_flat_dataclass_as_pytree(cls):
  def flatten(obj):
    children, aux_data = (), None
    for field in cls.__dataclass_fields__:
      children += (getattr(obj, field),)
    return (children, aux_data)

  def unflatten(aux_data, children):
    return cls(*children)

  register_pytree_node(cls, flatten, unflatten)
  return cls



#  Copyright 2025 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Integration tests for SFT trainer correctness."""

import pytest
import subprocess


@pytest.mark.integration_test
@pytest.mark.tpu_only
def test_maxtext_with_sft_in_trl():
  command = ["gsutil", "cp", "-r", "gs://maxtext-dataset/hf/llama3.1-tokenizer", "../assets/"]
  exit_code = subprocess.call(command)
  if exit_code != 0:
    raise ValueError(f"{command} failed with exit code: {exit_code}")

  command = [
      "python3",
      "MaxText/tests/sft_trainer_correctness.py",
      "--model-name=llama3.1-8b",
      "--tokenizer-path=assets/llama3.1-tokenizer",
      "--model-ckpt-path=gs://maxtext-model-checkpoints/llama3.1-8b/2025-01-23-19-04/scanned/0/items",
  ]

  subprocess.run(command, check=True, cwd="..")



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Attention kernel builder.

TODO: change to builder pattern instead of direct function call.
"""

import jax
from jax import numpy as jnp
from inference.kernel.attention import tpu

K_MASK = -2.3819763e38  # Set to a large negative number.


def vanilla_prefill_mha(q, k, v):
  """multi-head attention.

  q, k, v shape is (total_len, num_kv_heads, head_dim).
  output shape is (total_len, num_kv_heads * head_dim).
  """
  total_len = q.shape[0]
  num_kv_heads = q.shape[1]
  head_dim = q.shape[2]
  causal_mask = jnp.tril(jnp.ones(shape=(total_len, total_len))).astype(
      jnp.bool
  )
  mask = causal_mask[:, None, :]

  wei = jnp.einsum("tkh,skh->tks", q, k) / jnp.sqrt(head_dim)
  wei = jnp.where(mask, wei, K_MASK).astype(jnp.float32)
  wei = jax.nn.softmax(wei, axis=-1)
  out = jnp.einsum("tks,skh->tkh", wei, v)
  out = out.astype(q.dtype).reshape(total_len, num_kv_heads * head_dim)
  return out


def vanilla_prefill_mqa(q, k, v):
  """multi-query attention.

  q is (total_len, num_attn, head_dim).
  k/v shape is (total_len, 1, head_dim).
  output shape is (total_len, num_kv_heads * head_dim).
  """
  total_len = q.shape[0]
  head_dim = q.shape[2]
  k = jnp.squeeze(k, axis=1)
  v = jnp.squeeze(v, axis=1)
  causal_mask = jnp.tril(jnp.ones(shape=(total_len, total_len))).astype(
      jnp.bool
  )
  mask = causal_mask[:, None, :]

  wei = jnp.einsum("tah,sh->tas", q, k) / jnp.sqrt(head_dim)
  wei = jnp.where(mask, wei, K_MASK)
  wei = jax.nn.softmax(wei, axis=-1)
  out = jnp.einsum("tas,sh->tah", wei, v)
  out = out.astype(q.dtype).reshape((total_len, -1))
  return out


def vanilla_prefill_gqa(q, k, v):
  """group-query attention.

  q shape is (total_len, num_attn_heads, head_dim).
  k/v shape is (total_len, num_kv_heads, head_dim).
  output shape is (total_len, num_attn_heads * head_dim).
  """
  total_len, num_attn_heads, head_dim = q.shape
  num_kv_heads = k.shape[1]
  q = q.reshape(
      (total_len, num_kv_heads, num_attn_heads // num_kv_heads, head_dim)
  )
  causal_mask = jnp.tril(jnp.ones(shape=(total_len, total_len))).astype(
      jnp.bool
  )
  # padding_mask_row = jnp.where(jnp.arange(0, total_len) < len, 1, 0).astype(jnp.bool).reshape((-1, 1))
  # padding_mask_col = jnp.reshape(padding_mask_row, (1, -1))
  # padding_mask = jnp.logical_and(padding_mask_row, padding_mask_col)
  # mask = jnp.logical_and(padding_mask, causal_mask)[:, None, None, :]
  mask = causal_mask[:, None, None, :]

  wei = jnp.einsum("tkgh,skh->tkgs", q, k) / jnp.sqrt(head_dim)
  wei = jnp.where(mask, wei, K_MASK)
  wei = jax.nn.softmax(wei, axis=-1)
  out = jnp.einsum("tkgs,skh->tkgh", wei, v)
  out = out.astype(q.dtype).reshape((total_len, -1))
  return out


def chunked_prefill_attention(
    q, cache_k, cache_v, length, page_indices, accelerator="tpu"
):
  if accelerator == "tpu":
    return tpu.chunked_prefill_attention(
        q,
        cache_k,
        cache_v,
        length,
        page_indices,
    )
  else:
    raise NotImplementedError(f"not supported accelerate {accelerator}")


def decode_attention(q, cache_k, cache_v, pos, page_table, accelerator="tpu"):
  if accelerator == "tpu":
    # Heuristically set the pages per compute block.
    # TODO: tune the setup.
    pages_per_compute_block = 8
    _, _, _, head_dim = cache_k.shape
    q = q / jnp.sqrt(head_dim)
    seq_len = pos + 1

    output = tpu.paged_attention(
        q,
        cache_k,
        cache_v,
        seq_len,
        page_table,
        pages_per_compute_block=pages_per_compute_block,
    )
    return output

  else:
    raise NotImplementedError(f"not supported accelerate {accelerator}")



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""Transformer model definition."""
# pylint: disable=arguments-differ
# pylint: disable=no-name-in-module


from typing import Optional
from layers import quantizations
from layers import linears
from layers import initializers
import jax
from jax.ad_checkpoint import checkpoint_name
from jax.sharding import Mesh
from flax import linen as nn
import jax.numpy as jnp
from layers import attentions
from layers import embeddings
from layers import normalizations
from layers import models
import common_types
import max_logging

Array = common_types.Array
Config = common_types.Config
DType = common_types.DType
Mesh = common_types.Mesh
ScanIn = common_types.ScanIn

Embed = embeddings.Embed
Attention = attentions.Attention
RMSNorm = normalizations.RMSNorm
Quant = quantizations.AqtQuantization

# -----------------------------------------
# The Decoder Layer for DeepSeek v3
# -----------------------------------------
# Please note: DeepSeek V3 is not fully support at this moment


def self_attention_with_norm(inputs, cfg, mesh, quant, decoder_segment_ids, decoder_positions, deterministic, model_mode):
  # Normalization
  lnx_rms = models.RMSNorm(
      dtype=cfg.dtype,
      weight_dtype=cfg.weight_dtype,
      name="pre_self_attention_layer_norm",
      kernel_axes=("norm",),
      epsilon=cfg.normalization_layer_epsilon,
  )
  lnx = lnx_rms(inputs)
  lnx = nn.with_logical_constraint(lnx, ("activation_batch", "activation_norm_length", "activation_embed"))

  attention_layer = attentions.MLA(
      config=cfg,
      num_query_heads=cfg.num_query_heads,
      num_kv_heads=cfg.num_kv_heads,
      head_dim=cfg.head_dim,
      max_target_length=cfg.max_target_length,
      max_prefill_predict_length=cfg.max_prefill_predict_length,
      attention_kernel=cfg.attention,
      mesh=mesh,
      dtype=cfg.dtype,
      weight_dtype=cfg.weight_dtype,
      dropout_rate=cfg.dropout_rate,
      name="self_attention",
      quant=quant,
      kv_quant=quantizations.configure_kv_quant(cfg),
      q_lora_rank=cfg.q_lora_rank,
      kv_lora_rank=cfg.kv_lora_rank,
      qk_nope_head_dim=cfg.qk_nope_head_dim,
      qk_rope_head_dim=cfg.qk_rope_head_dim,
      v_head_dim=cfg.v_head_dim,
      max_position_embeddings=cfg.max_position_embeddings,
      original_max_position_embeddings=cfg.original_max_position_embeddings,
      mscale=cfg.mscale,
      rope_factor=cfg.rope_factor,
  )

  attention_lnx = attention_layer(
      lnx,
      lnx,
      decoder_positions,
      decoder_segment_ids=decoder_segment_ids,
      deterministic=deterministic,
      model_mode=model_mode,
  )

  attention_lnx = nn.with_logical_constraint(
      attention_lnx, ("activation_batch", "activation_norm_length", "activation_embed")
  )
  intermediate_inputs = inputs + attention_lnx

  # Normalization
  hidden_states = models.RMSNorm(
      dtype=cfg.dtype,
      weight_dtype=cfg.weight_dtype,
      name="post_self_attention_layer_norm",
      kernel_axes=("norm",),
      epsilon=cfg.normalization_layer_epsilon,
  )(intermediate_inputs)
  hidden_states = nn.with_logical_constraint(
      hidden_states, ("activation_batch", "activation_norm_length", "activation_embed")
  )
  return hidden_states, intermediate_inputs


def post_process(cfg, layer_output, sow):
  if cfg.record_internal_nn_metrics:
    sow("intermediates", "activation_mean", jnp.mean(layer_output))
    sow("intermediates", "activation_stdev", jnp.std(layer_output))
    sow(
        "intermediates",
        "activation_fraction_zero",
        jnp.sum(layer_output == 0) / jnp.size(layer_output),
    )

  if cfg.scan_layers:
    return layer_output, None
  else:
    return layer_output


class DeepSeekDenseLayer(nn.Module):
  """DeepSeek-style dense layer with Multi-Head Latent Attention."""

  config: models.Config
  mesh: Mesh
  quant: Optional[Quant] = None

  @nn.compact
  def __call__(
      self,
      inputs,
      decoder_segment_ids,
      decoder_positions,
      deterministic,
      model_mode,
      previous_chunk=None,
      page_state=None,
  ):
    cfg = self.config
    inputs = nn.with_logical_constraint(inputs, ("activation_batch", "activation_norm_length", "activation_embed"))
    inputs = checkpoint_name(inputs, "decoder_layer_input")

    hidden_states, intermediate_inputs = self_attention_with_norm(
        inputs, cfg, self.mesh, self.quant, decoder_segment_ids, decoder_positions, deterministic, model_mode
    )
    mlp_lnx = linears.MlpBlock(
        intermediate_dim=cfg.mlp_dim,
        activations=cfg.mlp_activations,
        intermediate_dropout_rate=cfg.dropout_rate,
        dtype=cfg.dtype,
        weight_dtype=cfg.weight_dtype,
        name="mlp",
        config=cfg,
        quant=self.quant,
    )(hidden_states, deterministic=deterministic)
    mlp_lnx = nn.with_logical_constraint(mlp_lnx, ("activation_batch", "activation_norm_length", "activation_embed"))

    layer_output = mlp_lnx + intermediate_inputs
    layer_output = nn.Dropout(rate=cfg.dropout_rate, broadcast_dims=(-2,))(layer_output, deterministic=deterministic)
    layer_output = nn.with_logical_constraint(
        layer_output,
        ("activation_batch", "activation_norm_length", "activation_embed"),
    )
    return post_process(cfg, layer_output, self.sow)


class DeepSeekMoELayer(nn.Module):
  """DeepSeek-style MoE layer with Multi-Head Latent Attention.
  Supports dropless and dropping base on configs.
  Uses a bias in routing instead of load balancing loss.
  """

  config: models.Config
  mesh: Mesh
  quant: Optional[Quant] = None

  @nn.compact
  def __call__(
      self,
      inputs,
      decoder_segment_ids,
      decoder_positions,
      deterministic,
      model_mode,
      previous_chunk=None,
      page_state=None,
  ):
    cfg = self.config
    inputs = nn.with_logical_constraint(inputs, ("activation_batch", "activation_norm_length", "activation_embed"))
    inputs = checkpoint_name(inputs, "decoder_layer_input")

    hidden_states, intermediate_inputs = self_attention_with_norm(
        inputs, self.config, self.mesh, self.quant, decoder_segment_ids, decoder_positions, deterministic, model_mode
    )

    mlp_lnx = linears.DeepSeekMoeBlock(
        config=cfg,
        mesh=self.mesh,
        kernel_init=initializers.nd_dense_init(1.0, "fan_in", "truncated_normal"),
        kernel_axes=("embed", None),
        dtype=cfg.dtype,
        weight_dtype=cfg.weight_dtype,
        quant=self.quant,
    )(hidden_states)
    mlp_lnx = nn.with_logical_constraint(mlp_lnx, ("activation_batch", "activation_norm_length", "activation_embed"))

    layer_output = mlp_lnx + intermediate_inputs
    layer_output = nn.Dropout(rate=cfg.dropout_rate, broadcast_dims=(-2,))(layer_output, deterministic=deterministic)
    layer_output = nn.with_logical_constraint(
        layer_output,
        ("activation_batch", "activation_norm_length", "activation_embed"),
    )
    return post_process(cfg, layer_output, self.sow)



#  Copyright 2025 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Page Managers for implementing paged attention in MaxText.

WARNING: THIS FILE IS A WORK IN PROGRESS.

This module provides the PageManager class and associated PageState dataclass for
managing the paged attention mechanism. The paging system allows efficient handling
of variable-length sequences by dividing the attention context into fixed-size pages,
similar to virtual memory systems.
"""

from typing import Optional, Tuple

import common_types
import jax
import jax.numpy as jnp
from flax import linen as nn
from flax import struct

Array = common_types.Array
DType = common_types.DType
AxisNames = common_types.AxisNames


@struct.dataclass
class PageState:
  """Represents the current state of the paging system.

  Attributes:
    page_status: Array indicating whether each page is in use (1) or free (0)
    page_map: Array mapping slots to their assigned pages
    sequence_lengths: Array containing the current length of each sequence
    num_pages_used: Array tracking how many pages each slot is using
    current_page: Array indicating the current active page for each slot
    current_page_position: Array tracking position within current pages
  """

  page_status: Array
  page_map: Array
  sequence_lengths: Array
  num_pages_used: Array
  current_page: Array
  current_page_position: Array


class PageManager(nn.Module):
  """Manages paged attention mechanism for efficient sequence processing.

  The PageManager implements a virtual memory-like system for attention, where the
  attention context is divided into fixed-size pages. This allows efficient handling
  of variable-length sequences and helps manage memory usage during inference.

  Attributes:
    num_pages: Total number of available pages in the system
    tokens_per_page: Number of tokens that can be stored in each page
    slots: Number of sequence slots available for parallel processing
    max_target_length: Maximum length of target sequences
    max_prefill_predict_length: Maximum length for prefill prediction
    max_pages_per_slot: Maximum number of pages that can be assigned to a slot
  """

  num_pages: int
  tokens_per_page: int
  slots: int
  max_target_length: int
  max_prefill_predict_length: int
  max_pages_per_slot: int

  def init_or_get_vars(self):
    """Initializes or retrieves the state variables for the paging system.

    Returns:
      Tuple of nn.Variable objects representing:
        - page_status: Status of each page (free/used)
        - page_map: Mapping between slots and their assigned pages
        - sequence_lengths: Length of sequence in each slot
        - num_pages_used: Number of pages used by each slot
        - current_page: Current active page for each slot
        - current_page_position: Position within current pages
    """
    page_status_var = self.variable(
        "cache", "page_status", nn.with_logical_partitioning(jnp.zeros, ("num_pages",)), (self.num_pages,), jnp.int32
    )
    page_map_var = self.variable(
        "cache",
        "page_map",
        nn.with_logical_partitioning(jnp.zeros, ("slots", "max_pages_per_slot")),
        (self.slots, self.max_pages_per_slot),
        jnp.int32,
    )
    sequence_lengths_var = self.variable(
        "cache", "sequence_lengths", nn.with_logical_partitioning(jnp.zeros, ("slots",)), (self.slots,), jnp.int32
    )
    num_pages_used_var = self.variable(
        "cache", "num_pages_used", nn.with_logical_partitioning(jnp.zeros, ("slots",)), (self.slots,), jnp.int32
    )
    current_page_var = self.variable(
        "cache", "current_page", nn.with_logical_partitioning(jnp.zeros, ("slots",)), (self.slots,), jnp.int32
    )
    current_page_position_var = self.variable(
        "cache", "current_page_position", nn.with_logical_partitioning(jnp.zeros, ("slots",)), (self.slots,), jnp.int32
    )

    return (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    )

  def release_slot_pages(
      self,
      slot: int,
      page_status_var: nn.Variable,
      page_map_var: nn.Variable,
      sequence_lengths_var: nn.Variable,
      num_pages_used_var: nn.Variable,
      current_page_var: nn.Variable,
      current_page_position_var: nn.Variable,
  ) -> Tuple:
    """Releases all pages assigned to a specific slot.

    This method frees up all pages currently assigned to the given slot,
    resetting their status and updating the page mapping accordingly.

    Args:
      slot: Integer identifying the slot to be released
      page_status_var: Variable tracking page usage status
      page_map_var: Variable mapping slots to pages
      sequence_lengths_var: Variable tracking sequence lengths
      num_pages_used_var: Variable tracking page usage counts
      current_page_var: Variable tracking current active pages
      current_page_position_var: Variable tracking positions in current pages

    Returns:
      Tuple of updated variables after releasing the slot's pages
    """
    page_status = page_status_var.value
    page_map = page_map_var.value
    sequence_lengths = sequence_lengths_var.value
    num_pages_used = num_pages_used_var.value
    current_page = current_page_var.value
    current_page_position = current_page_position_var.value

    def _release_page(i, state):
      page_map, page_status = state
      page_idx = page_map[slot][i]
      page_status = page_status.at[page_idx].set(0)
      page_map = page_map.at[slot, i].set(0)
      return page_map, page_status

    page_map, page_status = jax.lax.fori_loop(0, num_pages_used[slot], _release_page, (page_map, page_status))

    sequence_lengths = sequence_lengths.at[slot].set(0)
    num_pages_used = num_pages_used.at[slot].set(0)
    current_page = current_page.at[slot].set(0)
    current_page_position = current_page_position.at[slot].set(0)

    page_status_var.value = page_status
    page_map_var.value = page_map
    sequence_lengths_var.value = sequence_lengths
    num_pages_used_var.value = num_pages_used
    current_page_var.value = current_page
    current_page_position_var.value = current_page_position

    return (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    )

  def reserve_prefix_slot_pages(
      self,
      slot: int,
      true_length: int,
      page_status_var: nn.Variable,
      page_map_var: nn.Variable,
      sequence_lengths_var: nn.Variable,
      num_pages_used_var: nn.Variable,
      current_page_var: nn.Variable,
      current_page_position_var: nn.Variable,
  ) -> Tuple:
    """Reserves pages for a prefix sequence in the specified slot.

    This method allocates the necessary pages for a prefix sequence of given length,
    first releasing any existing pages assigned to the slot.

    Args:
      slot: Integer identifying the target slot
      true_length: Actual length of the prefix sequence
      page_status_var: Variable tracking page usage status
      page_map_var: Variable mapping slots to pages
      sequence_lengths_var: Variable tracking sequence lengths
      num_pages_used_var: Variable tracking page usage counts
      current_page_var: Variable tracking current active pages
      current_page_position_var: Variable tracking positions in current pages

    Returns:
      Tuple of updated variables after reserving pages for the prefix
    """
    (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    ) = self.release_slot_pages(
        slot,
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    )

    page_status = page_status_var.value
    page_map = page_map_var.value
    sequence_lengths = sequence_lengths_var.value
    num_pages_used = num_pages_used_var.value
    current_page = current_page_var.value
    current_page_position = current_page_position_var.value

    prefill_slot_num_pages = jnp.ceil(true_length / self.tokens_per_page).astype(jnp.int32)
    prefill_slot_page_slice_idx = jnp.where(true_length == 0, 0, (true_length - 1) % self.tokens_per_page)

    def _reserve_page(i, state):
      slot, page_map, page_status, current_page = state
      page_idx = jnp.where((page_status[1:] == 0), size=1)[0][0] + 1
      page_status = page_status.at[page_idx].set(1)
      page_map = page_map.at[slot, i].set(page_idx)
      current_page = current_page.at[slot].set(page_idx)
      return slot, page_map, page_status, current_page

    _, page_map, page_status, current_page = jax.lax.fori_loop(
        0, prefill_slot_num_pages, _reserve_page, (slot, page_map, page_status, current_page)
    )
    sequence_lengths = sequence_lengths.at[slot].set(true_length)
    num_pages_used = num_pages_used.at[slot].set(prefill_slot_num_pages)
    current_page_position = current_page_position.at[slot].set(prefill_slot_page_slice_idx)

    page_status_var.value = page_status
    page_map_var.value = page_map
    sequence_lengths_var.value = sequence_lengths
    num_pages_used_var.value = num_pages_used
    current_page_var.value = current_page
    current_page_position_var.value = current_page_position

    return (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    )

  def reserve_decode_step_pages(
      self,
      page_status_var: nn.Variable,
      page_map_var: nn.Variable,
      sequence_lengths_var: nn.Variable,
      num_pages_used_var: nn.Variable,
      current_page_var: nn.Variable,
      current_page_position_var: nn.Variable,
  ) -> Tuple:
    """Reserves additional pages needed for a decoding step.

    This method allocates new pages as needed when sequences grow during
    autoregressive decoding, ensuring each active slot has sufficient pages
    for its sequence.

    Args:
      page_status_var: Variable tracking page usage status
      page_map_var: Variable mapping slots to pages
      sequence_lengths_var: Variable tracking sequence lengths
      num_pages_used_var: Variable tracking page usage counts
      current_page_var: Variable tracking current active pages
      current_page_position_var: Variable tracking positions in current pages

    Returns:
      Tuple of updated variables after reserving pages for the decode step
    """
    page_status = page_status_var.value
    page_map = page_map_var.value
    sequence_lengths = sequence_lengths_var.value
    num_pages_used = num_pages_used_var.value
    current_page = current_page_var.value
    current_page_position = current_page_position_var.value

    sequence_lengths_step = jnp.logical_and(jnp.ones(sequence_lengths.shape, dtype=jnp.int32), sequence_lengths).astype(
        jnp.int32
    )

    sequence_lengths += sequence_lengths_step

    current_num_pages_used = num_pages_used
    num_pages_used = jnp.ceil(sequence_lengths / self.tokens_per_page).astype(jnp.int32)

    current_page_position = jnp.where(sequence_lengths == 0, 0, (sequence_lengths - 1) % self.tokens_per_page)
    seq_new_page = num_pages_used - current_num_pages_used

    updating_slots = jnp.where((seq_new_page > 0), size=self.slots)[0]

    def _reserve_page(i, state):
      page_map, page_status, current_page, updating_slots = state
      slot = jax.lax.dynamic_index_in_dim(updating_slots, i, axis=0, keepdims=False)
      page_idx = jnp.where((page_status[1:] == 0), size=1)[0][0] + 1
      page_status = page_status.at[page_idx].set(1)
      page_map = page_map.at[slot, num_pages_used[slot] - 1].set(page_idx)
      current_page = current_page.at[slot].set(page_idx)
      return page_map, page_status, current_page, updating_slots

    page_map, page_status, current_page, _ = jax.lax.fori_loop(
        0, jnp.count_nonzero(seq_new_page), _reserve_page, (page_map, page_status, current_page, updating_slots)
    )

    page_status_var.value = page_status
    page_map_var.value = page_map
    sequence_lengths_var.value = sequence_lengths
    num_pages_used_var.value = num_pages_used
    current_page_var.value = current_page
    current_page_position_var.value = current_page_position

    return (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    )

  @nn.compact
  def __call__(
      self, model_mode: Optional[str] = None, slot: Optional[int] = None, true_length: Optional[int] = None
  ) -> PageState:

    (
        page_status_var,
        page_map_var,
        sequence_lengths_var,
        num_pages_used_var,
        current_page_var,
        current_page_position_var,
    ) = self.init_or_get_vars()

    if model_mode == common_types.MODEL_MODE_PREFILL and self.is_mutable_collection("params"):
      return PageState(
          page_status_var.value,
          page_map_var.value,
          sequence_lengths_var.value,
          num_pages_used_var.value,
          current_page_var.value,
          current_page_position_var.value,
      )
    elif model_mode == common_types.MODEL_MODE_PREFILL and slot is None and true_length is None:
      return PageState(
          page_status_var.value,
          page_map_var.value,
          sequence_lengths_var.value,
          num_pages_used_var.value,
          current_page_var.value,
          current_page_position_var.value,
      )
    elif model_mode == common_types.MODEL_MODE_PREFILL:
      self.reserve_prefix_slot_pages(
          slot,
          true_length,
          page_status_var,
          page_map_var,
          sequence_lengths_var,
          num_pages_used_var,
          current_page_var,
          current_page_position_var,
      )
    elif model_mode == common_types.MODEL_MODE_INSERT:
      self.reserve_prefix_slot_pages(
          slot,
          true_length,
          page_status_var,
          page_map_var,
          sequence_lengths_var,
          num_pages_used_var,
          current_page_var,
          current_page_position_var,
      )
    elif model_mode == common_types.MODEL_MODE_AUTOREGRESSIVE:
      self.reserve_decode_step_pages(
          page_status_var,
          page_map_var,
          sequence_lengths_var,
          num_pages_used_var,
          current_page_var,
          current_page_position_var,
      )

    return PageState(
        page_status_var.value,
        page_map_var.value,
        sequence_lengths_var.value,
        num_pages_used_var.value,
        current_page_var.value,
        current_page_position_var.value,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for metrics."""

import unittest

import benchmarks.metrics as metrics
import datetime
import re


class TestEventMetric(unittest.TestCase):
  """ "Tests for event metric (i.e. distribution)."""

  def setUp(self):
    self._metric = metrics.EventMetric(
        "requestLatency", "Latency of requests", "ms"
    )

  def test_record_adds_a_data_event(self):
    m = self._metric
    m.record(1.0)
    data_points = m.data()
    self.assertEqual(1, len(data_points))
    self.assertEqual(1.0, data_points[0])

  def test_percentile_returns_correct_percentile(self):
    m = self._metric
    n = 11
    for i in range(0, n):
      m.record(i)
    self.assertEqual(m.percentile(50), 5)
    self.assertEqual(m.percentile(90), 9)
    self.assertEqual(m.percentile(100), 10)

  def test_mean_returns_correct_mean_value(self):
    m = self._metric
    n = 3
    for i in range(0, n):
      m.record(i)
    self.assertEqual(sum(range(0, n)) / n, m.mean())

  def test_distribution_summary_str_returns_expected_str(self):
    m = self._metric
    n = 100
    for i in range(0, n):
      m.record(i)
    summary = m.distribution_summary_str()
    self.assertTrue(re.search(r"Mean requestLatency", summary))
    self.assertTrue(re.search(r"Median requestLatency", summary))
    self.assertTrue(re.search(r"P99 requestLatency", summary))

  def test_distribution_summary_dict_returns_expected_dict(self):
    m = self._metric
    n = 100
    for i in range(0, n):
      m.record(i)
    summary = m.distribution_summary_dict()
    self.assertIn("mean_requestLatency_ms", summary)
    self.assertIn("median_requestLatency_ms", summary)
    self.assertIn("p99_requestLatency_ms", summary)


class TestCounterMetric(unittest.TestCase):
  """Tests for counter metric (i.e. monotonically increasing counter)."""

  def setUp(self):
    self._counter = metrics.CounterMetric(
        "RequestCompleteCount", "Number of completed requests"
    )

  def test_increment_increases_total_count(self):
    m = self._counter
    old_total_cnt = m.total_count()
    m.increment()
    new_total_cnt = m.total_count()
    self.assertEqual(1, new_total_cnt - old_total_cnt)

  def test_increment_within_same_second_update_counts_for_the_second(self):
    """Test to ensure one entry used to cumulate counts within a second"""
    m = self._counter
    timestamp = datetime.datetime.strptime(
        "2025-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"
    )
    m.increment(1, timestamp)
    m.increment(2, timestamp)
    data = m.data()
    self.assertEqual(1, len(data.keys()))
    self.assertEqual(3, data[timestamp])

  def test_increment_at_different_seconds_creates_separate_entries(self):
    """Test to ensure separate entries used for different seconds"""
    m = self._counter
    timestamp_first = datetime.datetime.strptime(
        "2025-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"
    )
    m.increment(1, timestamp_first)
    timestamp_second = datetime.datetime.strptime(
        "2025-01-01 00:00:01", "%Y-%m-%d %H:%M:%S"
    )
    m.increment(2, timestamp_second)
    data = m.data()
    self.assertEqual(2, len(data.keys()))
    self.assertEqual(1, data[timestamp_first])
    self.assertEqual(2, data[timestamp_second])

  def test_rate_returns_expected(self):
    m = self._counter
    n = 10
    start_time = datetime.datetime.strptime(
        "2025-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"
    )
    delta_time_sec = 1
    for i in range(0, n):
      m.increment(
          1, start_time + datetime.timedelta(seconds=delta_time_sec * i)
      )
    # n counts across n seconds, thus rate = 1
    self.assertEqual(1, m.rate())

  def test_rate_over_window_returns_expected(self):
    m = self._counter
    n = 10
    start_time = datetime.datetime.strptime(
        "2025-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"
    )
    delta_time_sec = 1
    for i in range(0, n):
      m.increment(
          1, start_time + datetime.timedelta(seconds=delta_time_sec * i)
      )

    rates_with_timestamp = m.rate_over_window(window_size_sec=5)

    rates = [rate for timestamp, rate in rates_with_timestamp]
    # 10 seconds with 1 count in each second. One rate per window_size_sec=5 sec
    # so rate = 1 for [0, 5) and rate = 1 [5, 10)
    self.assertEqual([1, 1], rates)

  def test_rate_over_window_to_csv_returns_correct(self):
    m = self._counter
    n = 10
    start_time = datetime.datetime.strptime(
        "2025-01-01 00:00:00", "%Y-%m-%d %H:%M:%S"
    )
    delta_time_sec = 1
    for i in range(0, n):
      m.increment(
          1, start_time + datetime.timedelta(seconds=delta_time_sec * i)
      )

    csv_output = m.rate_over_window_to_csv(window_size_sec=5)

    rows = csv_output.split("\n")
    self.assertEqual(2, len(rows))
    expected_timestamps = "TimeStamp,2025-01-01 00:00:00,2025-01-01 00:00:05"
    got_timestamps = rows[0]
    self.assertEqual(expected_timestamps, got_timestamps)
    expected_values = "Value,1.00,1.00"
    got_values = rows[1]
    self.assertEqual(expected_values, got_values)


if __name__ == "__main__":
  unittest.main()



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from absl.testing import absltest
import numpy as np
from jax import numpy as jnp
from inference.nn import Module, Parameter


class ModuleTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.w0, self.w1, self.w2, self.w3 = (
        jnp.ones((1,)),
        jnp.ones((2,)),
        jnp.ones((3,)),
        jnp.ones((4,)),
    )
    self.parent_module = Module()
    self.parent_module.w0 = Parameter(self.w0)
    self.h1_child_0_module = Module()
    self.h1_child_0_module.w1 = Parameter(self.w1)
    self.h1_child_1_module = Module()
    self.h1_child_1_module.w2 = Parameter(self.w2)
    self.h2_child_0_module = Module()
    self.h2_child_0_module.w3 = Parameter(self.w3)

    self.parent_module.child0 = self.h1_child_0_module
    self.parent_module.child1 = self.h1_child_1_module
    self.h1_child_0_module.child0 = self.h2_child_0_module

  def test_random_code_initialize(self):
    self.parent_module.init_weights()
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        self.w0,
        self.parent_module.w0.value,
    )
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        self.w1,
        self.h1_child_0_module.w1.value,
    )
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        self.w2,
        self.h1_child_1_module.w2.value,
    )
    np.testing.assert_raises(
        AssertionError,
        np.testing.assert_array_equal,
        self.w3,
        self.h2_child_0_module.w3.value,
    )

  def test_load_weights_dict(self):
    parent_weight_dict = {
        "w0": jnp.ones((1,)),
        "child0": {
            "w1": jnp.ones((2,)),
            "child0": {
                "w3": jnp.ones((4,)),
            },
        },
        "child1": {
            "w2": jnp.ones((3,)),
        },
    }

    self.parent_module.load_weights_dict(parent_weight_dict)

    np.testing.assert_array_equal(self.parent_module.w0, self.w0)
    np.testing.assert_array_equal(self.h1_child_0_module.w1, self.w1)
    np.testing.assert_array_equal(self.h1_child_1_module.w2, self.w2)
    np.testing.assert_array_equal(self.h2_child_0_module.w3, self.w3)

  def test_load_weights_dict_error(self):
    partial_parent_weight_dict = {
        "w0": jnp.zeros((1,)),
        "child0": {
            "w1": jnp.zeros((2,)),
            "child0": {
                "w3": jnp.zeros((4,)),
            },
        },
    }
    child1_weight_dict = {
        "w2": jnp.zeros((2,)),
        "wrong_weight_not_load": jnp.zeros((2,)),
    }

    self.parent_module.load_weights_dict(partial_parent_weight_dict)
    self.h1_child_1_module.load_weights_dict(child1_weight_dict)

    np.testing.assert_array_equal(self.parent_module.w0, 0)
    np.testing.assert_array_equal(self.h1_child_0_module.w1, 0)
    np.testing.assert_array_equal(self.h1_child_1_module.w2, 0)
    np.testing.assert_array_equal(self.h2_child_0_module.w3, 0)

    assert not self.h1_child_1_module.wrong_weight_not_load


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Training loop for Supervised Fine-Tuning (SFT)."

import datetime
import os
import sys
import queue

from typing import Sequence
from absl import app
from flax.linen import partitioning as nn_partitioning
import jax
import numpy as np

import checkpointing
import max_utils
import maxtext_utils
import max_logging
import profiler
import pyconfig
import tensorflow as tf

from input_pipeline.input_pipeline_interface import create_data_iterator
from gcp_workload_monitor import GCPWorkloadMonitor
from metric_logger import MetricLogger
from train import (
    check_example_batch,
    create_goodput_recorder,
    eval_step,
    EPS,
    get_first_step,
    load_next_batch,
    record_goodput,
    record_scalar_metrics,
    save_checkpoint,
    setup_mesh_and_model,
    train_step,
    validate_train_config,
)
from ml_goodput_measurement import monitoring


def setup_train_loop(config):
  """Set up prerequisites for the training loop -
    checkpoint_manager, PRNG keys, Mesh, Model and optimizer.
    Set up data iterator and tokenizer, initialize the model.

  Args:
      config

  Returns:
      init_rng:
      writer: Summary writer for tensorboard
      checkpoint_manager: Orbax checkpointer
      state_mesh_annotations: the mesh annotations for the train state
      model:
      mesh:
      learning_rate_schedule:
      data_iterator:
      state: the initialized train state
  """
  recorder = create_goodput_recorder(config)

  record_goodput(recorder, config, recorder.record_tpu_init_start_time if recorder else None)
  init_rng, writer, checkpoint_manager, mesh, model, learning_rate_schedule, tx = setup_mesh_and_model(config)
  record_goodput(recorder, config, recorder.record_tpu_init_end_time if recorder else None)

  record_goodput(recorder, config, recorder.record_training_preparation_start_time if recorder else None)
  data_iterator, eval_data_iterator = create_data_iterator(config, mesh)
  state, _, state_mesh_shardings, data_iterator = max_utils.setup_training_state(
      model, data_iterator, tx, config, init_rng, mesh, checkpoint_manager
  )
  if not config.using_pipeline_parallelism:
    # The vocab tensor(s) of shape [vocab, embed] (and transpose) are not sharded by stage
    maxtext_utils.assert_params_sufficiently_sharded(state.params, mesh, config.sharding_tolerance)
  record_goodput(recorder, config, recorder.record_training_preparation_end_time if recorder else None)
  return (
      init_rng,
      writer,
      checkpoint_manager,
      state_mesh_shardings,
      model,
      mesh,
      learning_rate_schedule,
      data_iterator,
      eval_data_iterator,
      state,
  )


def train_loop(config, state=None):
  """Main training loop for SFT."""
  if not config.use_sft:
    raise TypeError("Set use_sft to True to run Supervised Fine Tuning.")

  # Create a GoodputRecorder to log information
  recorder = create_goodput_recorder(config)
  record_goodput(recorder, config, recorder.record_job_start_time if recorder else None)

  (
      init_rng,
      writer,
      checkpoint_manager,
      state_mesh_shardings,
      model,
      mesh,
      learning_rate_schedule,
      data_iterator,
      eval_data_iterator,
      state,
  ) = setup_train_loop(config)

  # pylint: disable=line-too-long
  (
      functional_train,
      in_shard_train,
      out_shard_train,
      static_argnums_train,
      donate_argnums_train,
  ) = maxtext_utils.get_functional_train_with_signature(train_step, mesh, state_mesh_shardings, model, config)

  if eval_data_iterator:
    # pylint: disable=line-too-long
    (
        functional_eval,
        in_shard_eval,
        out_shard_eval,
        static_argnums_eval,
        donate_argnums_eval,
    ) = maxtext_utils.get_functional_eval_with_signature(eval_step, mesh, state_mesh_shardings, model, config)

  num_model_parameters = max_utils.calculate_num_params_from_pytree(state.params)
  max_logging.log(f"number parameters: {num_model_parameters/1e9:.3f} billion")
  per_device_tflops, _, _ = maxtext_utils.calculate_tflops_training_per_device(config)
  per_device_tokens = maxtext_utils.calculate_tokens_training_per_device(config)

  # Write train config params, num model params, and XLA flags to tensorboard
  max_utils.add_text_to_summary_writer("num_model_parameters", str(num_model_parameters), writer)
  max_utils.add_text_to_summary_writer("libtpu_init_args", os.environ["LIBTPU_INIT_ARGS"], writer)
  max_utils.add_config_to_summary_writer(config, writer)

  # Define the compilation of functional_train, either by loading the compiled version or wrapping a new one in a jit
  if config.compiled_trainstep_file != "":
    print("Loading the compiled function...", flush=True)
    # Need to pass train signature and state to determine i/o shapes of train_state for now.
    p_train_step = maxtext_utils.load_compiled(config, functional_train, state)
    # TODO: p_eval_step is not yet supported in load_compiled
    p_eval_step = None
    print("Loaded compiled function!", flush=True)
  else:
    p_train_step = jax.jit(
        functional_train,
        in_shardings=in_shard_train,
        out_shardings=out_shard_train,
        static_argnums=static_argnums_train,
        donate_argnums=donate_argnums_train,
    )

    p_eval_step = None
    if eval_data_iterator:
      p_eval_step = jax.jit(
          functional_eval,
          in_shardings=in_shard_eval,
          out_shardings=out_shard_eval,
          static_argnums=static_argnums_eval,
          donate_argnums=donate_argnums_eval,
      )

  running_gcs_metrics = [] if config.gcs_metrics else None

  start_step = get_first_step(state)  # this is the start_step for training
  prof = profiler.Profiler(config, offset_step=start_step)
  first_profiling_step = prof.start_initial_profile_step
  if config.profiler != "" and first_profiling_step >= config.steps:
    raise ValueError("Profiling requested but initial profiling step set past training final step")
  last_profiling_step = prof.finished_initial_profile_step

  example_batch = None
  last_step_completion = datetime.datetime.now()

  performance_metric_queue = None
  if config.report_heartbeat_metric_for_gcp_monitoring or config.report_performance_metric_for_gcp_monitoring:
    gcp_workload_monitor = GCPWorkloadMonitor(config.run_name)
    if config.report_heartbeat_metric_for_gcp_monitoring:
      gcp_workload_monitor.start_heartbeat_reporting_thread(config.heartbeat_reporting_interval_in_seconds)
    if config.report_performance_metric_for_gcp_monitoring:
      performance_metric_queue = queue.Queue()
      gcp_workload_monitor.start_performance_reporting_thread(performance_metric_queue)

  metric_logger = MetricLogger(writer, config)
  for step in np.arange(start_step, config.steps):
    if step == first_profiling_step or prof.should_activate_periodic_profile(step):
      optional_postfix = f"step_{step}" if config.profile_periodically_period > 0 else ""
      prof.activate(blocking_object=state, optional_postfix=optional_postfix)

    with jax.profiler.StepTraceAnnotation("train", step_num=step):
      record_goodput(recorder, config, recorder.record_data_loading_start_time if recorder else None)
      try:
        example_batch = load_next_batch(data_iterator, example_batch, config)
      except Exception as e:  # pylint: disable=broad-except
        max_logging.log(f"load_next_batch failed, you may run out of data. Error message: {e}")
        if checkpoint_manager is not None:
          if save_checkpoint(
              checkpoint_manager, int(state.step), state, config.dataset_type, data_iterator, config, force=True
          ):
            checkpointing.print_save_message(state.step, config.async_checkpointing)
        break
      record_goodput(recorder, config, recorder.record_data_loading_end_time if recorder else None)
      check_example_batch(config, example_batch=example_batch)
      # pylint: disable=not-callable
      nextrng = jax.jit(jax.random.fold_in)(init_rng, step)
      record_goodput(recorder, config, recorder.record_step_start_time if recorder else None, step)
      with mesh, nn_partitioning.axis_rules(config.logical_axis_rules):
        state, metrics = p_train_step(state, example_batch, nextrng)

    step_time_delta = datetime.datetime.now() - last_step_completion
    last_step_completion = datetime.datetime.now()
    record_scalar_metrics(metrics, step_time_delta, per_device_tflops, learning_rate_schedule(step), per_device_tokens)
    if performance_metric_queue:
      performance_metric_queue.put(step_time_delta.total_seconds())

    if checkpoint_manager is not None:
      if save_checkpoint(checkpoint_manager, int(step), state, config.dataset_type, data_iterator, config):
        checkpointing.print_save_message(step, config.async_checkpointing)

      # Upon preemption, exit when and only when all ongoing saves are complete.
      if checkpoint_manager.reached_preemption(step):
        checkpoint_manager.wait_until_finished()
        sys.exit()

    metric_logger.write_metrics(running_gcs_metrics, metrics, step)

    if config.dump_hlo and step == start_step:
      jax.block_until_ready(state)  # Ensure compilation has finished.
      max_utils.upload_dump(
          config.dump_hlo_local_dir,
          config.dump_hlo_gcs_dir,
          module_name=config.dump_hlo_module_name,
          delete_local_after=config.dump_hlo_delete_local_after,
          all_host_upload=config.dump_hlo_upload_all,
      )

    if config.eval_interval > 0 and step > start_step and (step + 1) % config.eval_interval == 0:
      assert eval_data_iterator
      cumulative_eval_metrics = {
          "scalar": {
              "eval/total_loss": 0.0,
              "eval/total_weights": 0.0,
              "eval/avg_loss": 0.0,
              "eval/moe_lb_loss": 0.0,
          }
      }
      eval_step_count = 0
      # pylint: disable=not-callable
      for eval_batch in eval_data_iterator:
        if config.eval_steps > 0 and eval_step_count >= config.eval_steps:
          break
        with mesh, nn_partitioning.axis_rules(config.logical_axis_rules):
          eval_metrics = p_eval_step(state, eval_batch, nextrng)
        cumulative_eval_metrics["scalar"]["eval/total_loss"] += float(eval_metrics["scalar"]["evaluation/total_loss"])
        cumulative_eval_metrics["scalar"]["eval/total_weights"] += float(eval_metrics["scalar"]["evaluation/total_weights"])
        cumulative_eval_metrics["scalar"]["eval/moe_lb_loss"] += float(eval_metrics["scalar"]["evaluation/moe_lb_loss"])
        max_logging.log(f"Completed eval step {eval_step_count}")
        eval_step_count += 1
      eval_loss = cumulative_eval_metrics["scalar"]["eval/total_loss"] / (
          cumulative_eval_metrics["scalar"]["eval/total_weights"] + EPS
      )
      cumulative_eval_metrics["scalar"]["eval/avg_loss"] = eval_loss
      cumulative_eval_metrics["scalar"]["eval/avg_moe_lb_loss"] = (
          cumulative_eval_metrics["scalar"]["eval/moe_lb_loss"] / eval_step_count
      )
      metric_logger.write_metrics(running_gcs_metrics, cumulative_eval_metrics, step, is_training=False)
      max_logging.log(
          f"average loss after {step=}: {eval_step_count=}, {eval_loss=},"
          f" total_weights={cumulative_eval_metrics['scalar']['eval/total_weights']}"
      )
      if eval_loss <= config.target_eval_loss:
        max_logging.log(f"Early stop and exit loop after reaching {config.target_eval_loss=}")
        prof.deactivate()
        break

    if step == last_profiling_step or prof.should_deactivate_periodic_profile(step):
      prof.deactivate(blocking_object=state)

    if step == start_step:
      max_utils.print_mem_stats("After params initialized")

  if checkpoint_manager is not None:
    checkpoint_manager.wait_until_finished()
  metric_logger.write_metrics(running_gcs_metrics, metrics, config.steps - 1)  # final step metrics
  max_utils.close_summary_writer(writer)
  record_goodput(recorder, config, recorder.record_job_end_time if recorder else None)

  with mesh, nn_partitioning.axis_rules(config.logical_axis_rules):
    compiled = p_train_step.lower(state, example_batch, nextrng).compile()
    compiled_stats = compiled.memory_analysis()
    if compiled_stats is not None:
      max_logging.log(
          f"Output size: {compiled_stats.output_size_in_bytes}, "
          f"temp size: {compiled_stats.temp_size_in_bytes}, "
          f"argument size: {compiled_stats.argument_size_in_bytes}, "
          f"host temp size: {compiled_stats.host_temp_size_in_bytes}, in bytes."
      )
  return state


def main(argv: Sequence[str]) -> None:
  jax.config.update("jax_default_prng_impl", "unsafe_rbg")
  # TF allocates extraneous GPU memory when using TFDS data
  # this leads to CUDA OOMs. WAR for now is to hide GPUs from TF
  tf.config.set_visible_devices([], "GPU")
  os.environ["TF_CPP_MIN_LOG_LEVEL"] = "0"
  if "xla_tpu_spmd_rng_bit_generator_unsafe" not in os.environ.get("LIBTPU_INIT_ARGS", ""):
    os.environ["LIBTPU_INIT_ARGS"] = os.environ.get("LIBTPU_INIT_ARGS", "") + " --xla_tpu_spmd_rng_bit_generator_unsafe=true"
  config = pyconfig.initialize(argv)
  max_utils.print_system_information()
  validate_train_config(config)
  os.environ["TFDS_DATA_DIR"] = config.dataset_path

  if config.monitor_goodput and jax.process_index() == 0:
    logger_name = f"goodput_{config.run_name}"
    goodput_monitor = monitoring.GoodputMonitor(
        job_name=config.run_name,
        logger_name=logger_name,
        tensorboard_dir=config.tensorboard_dir,
        upload_interval=config.goodput_upload_interval_seconds,
        monitoring_enabled=True,
        pathway_enabled=config.enable_pathways_goodput,
        include_badput_breakdown=True,
        include_step_deviation=config.monitor_step_time_deviation,
        step_deviation_interval_seconds=config.step_deviation_interval_seconds,
    )
    goodput_monitor.start_goodput_uploader()
    max_logging.log("Started Goodput upload to Tensorboard in the background!")
    if config.monitor_step_time_deviation:
      goodput_monitor.start_step_deviation_uploader()
      max_logging.log("Started step time deviation upload to Tensorboard in the background!")

  train_loop(config)


if __name__ == "__main__":
  app.run(main)



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""mesh module"""

import math
from collections.abc import Iterable
from typing import Sequence
import jax
from jax.sharding import Mesh
from jax.experimental import mesh_utils as jax_mesh_utils
from jaxlib.xla_client import Device
import numpy as np
from inference.parallel.config import ParallelAxis
from inference.parallel.device import platform


def create_device_mesh(
    devices: Sequence[Device],
    shape: tuple[int],
) -> Mesh:
  """Create a powerful mesh given the devices and shape.

  For fully connected topology, the devices topology defined in the mesh
  usually not affect the performance.
  For other cases, the devices topology defined in the mesh
  will affect the performance. (it depends on the collective algorithm
  implementation.)
  """
  assert devices, "no devices is provided for mesh creation"
  assert len(devices) == math.prod(shape)
  axis_names = (
      ParallelAxis.X.name,
      ParallelAxis.Y.name,
  )
  assert len(shape) == len(axis_names)

  p = platform()
  if p == "gpu" or p == "cpu":
    devices = jax_mesh_utils.create_device_mesh(
        mesh_shape=shape,
        devices=devices,
        allow_split_physical_axes=True,
    )
    return Mesh(devices=devices, axis_names=axis_names)
  else:
    assert p == "tpu"
    # TODO: Figure out a general method.
    # Current mesh builder is very limited.
    # only support (2,x) underlying topology shape to .
    # form a 1D ring for the devices.
    devices = devices[::2] + devices[1::2][::-1]
    devices = np.reshape(devices, shape)
    return Mesh(devices=devices, axis_names=axis_names)


def get_num_partitions(axis_names):
  """Get the total number of partitions across the axis.
  Args:
      axis_names: the name of the axis where the partition is
          relative to. If the number of axis is greater than 1,
          the axis names need to follow the "major to minor" order
          as defined in the mesh setup for consistency.
  """
  if not isinstance(axis_names, Iterable):
    return jax.lax.psum(1, axis_name=axis_names)
  product = 1
  for axis in axis_names[::-1]:
    product = product * jax.lax.psum(1, axis_name=axis)
  return product


def get_partition_index(axis_names):
  """Get the partition index across the axis for the device.
  Args:
      axis_names: the names of the axis where the partition index is
          relative to. If the number of axis is greater than 1,
          the axis names need to follow the "major to minor" order
          as defined in the mesh setup for consistency.
  """
  if not isinstance(axis_names, Iterable):
    return jax.lax.axis_index(axis_name=axis_names)
  cur_idx = 0
  multiplicand = 1
  for axis in axis_names[::-1]:
    cur_idx += jax.lax.axis_index(axis) * multiplicand
    multiplicand = multiplicand * jax.lax.psum(1, axis_name=axis)
  return cur_idx



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .module import Module, ModuleList, Model, CausalLM
from .parameter import Parameter
from .linear import Linear
from .embedding import Embedding, apply_rope_embedding
from .norm import RMSNorm
from .attention import AttentionOps, AttentionMetadata, KVCache



"""
Copyright 2025 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Optional, Union, List, Dict
import numpy as np
from transformers import AutoTokenizer
from kithara.model.hf_compatibility import get_model_name_from_preset_handle
from kithara.model.maxtext.conversion_utils import MaxTextConversionMixin
from kithara.model.maxtext.ckpt_compatibility import (
    save_maxtext_model_in_hf_format,
    load_hf_weights_into_maxtext_model,
)
from kithara.model import (
    Model,
    set_precision,
    set_global_model_implementation_type,
    ModelImplementationType,
    supported_models,
)
from kithara.model.maxtext.inference_engine import MaxtextInferenceEngine as MaxEngine
from tqdm import tqdm


class MaxTextModel(Model, MaxTextConversionMixin):
    """
    MaxTextModel is class that represents a MaxText model via the
    Kithara.Model interface. It is a thin wrapper around the underlying
    MaxText model instance.

    Methods
    -------
    from_random: Create a randomly initialized MaxText model with the
        given configuration.
    from_preset: Create a MaxText model initialized with weights from
        HuggingFace Hub.
    generate: Generate text based on the input tokens, with an option to
        stop at specific token IDs.
    save_in_hf_format: Save the MaxText model in HuggingFace format.
    """

    def __init__(
        self,
        model,
        sharding_strategy,
        model_name: str,
        precision: str,
        scan_layers: bool,
        maxtext_config: dict,
    ):
        """Initialize a MaxTextModel instance.

        Args:
            model: The underlying MaxText model instance.
            sharding_strategy: Strategy for sharding the model across devices.
            model_name (str): Name of the MaxText model.
            precision (str): Precision mode used for computations.
            scan_layers (bool): Whether scan layers are used for memory efficiency.
            maxtext_config (dict): Configuration parameters for the MaxText model.
        """
        super().__init__(
            model,
            sharding_strategy,
            model_name,
            precision,
            scan_layers,
        )

        self.maxtext_config = maxtext_config
        self.maxengine = None
        self.decode_state = None

    @classmethod
    def from_random(
        cls,
        model_name: str,
        seq_len: int = 2048,
        per_device_batch_size: int = 1,
        precision: str = "mixed_bfloat16",
        scan_layers: bool = True,
        max_prefill_predict_length: int = "auto",
        maxtext_config_args: Optional[dict] = None,
    ) -> "MaxTextModel":
        """Create a randomly initialized MaxText model with the given configuration.

        Args:
            model_name (str): Name of the MaxText model configuration to use.
            seq_len (int, optional): Maximum sequence length. Defaults to 2048.
            per_device_batch_size (int, optional): Batch size per device.
                Defaults to 1.
            precision (str, optional): Precision mode for computations.
                Defaults to "mixed_bfloat16".
            scan_layers (bool, optional): Whether to use scan layers for memory efficiency.
                Defaults to True.
            max_prefill_predict_length (int, optional): The maximum length of prompt tokens supported at
                inferenece time. The number of generated tokens is capped to max_seq_len -
                max_prefill_length. This number is required by the inference server. It is
                defaulted to "auto" which sets to `seq_len//2`. You should set it a larger number
                if you are running inference with this model on larger prompts; or set it to a
                smaller number if you are working with short prompts and what to generate more tokens.
            maxtext_config_args (Optional[dict], optional): Additional configuration arguments.
                Defaults to None.

        Returns:
            MaxTextModel: A new instance of MaxTextModel with random initialization.
        """
        set_global_model_implementation_type(ModelImplementationType.MAXTEXT)

        set_precision(precision)
        weight_dtype = cls._weight_dtype(precision)
        activation_dtype = cls._activation_dtype(precision)

        if max_prefill_predict_length == "auto":
            max_prefill_predict_length = seq_len // 2

        maxtext_config, sharding_strategy, model = cls.initialize_random_maxtext_model(
            model_name,
            seq_len,
            per_device_batch_size,
            weight_dtype,
            activation_dtype,
            scan_layers,
            max_prefill_predict_length,
            maxtext_config_args,
        )

        return cls(
            model,
            sharding_strategy,
            model_name=model_name,
            precision=precision,
            scan_layers=scan_layers,
            maxtext_config=maxtext_config,
        )

    @classmethod
    def from_preset(
        cls,
        preset_handle: str,
        seq_len: int = 2048,
        per_device_batch_size: int = 1,
        precision: str = "mixed_bfloat16",
        scan_layers: bool = "auto",
        max_prefill_predict_length: int = "auto",
        maxtext_config_args: Optional[dict] = None,
    ) -> "MaxTextModel":
        """Create a MaxText model initialized with weights from HuggingFace Hub.

        Args:
            preset_handle (str): HuggingFace model identifier. This could be a
                HuggingFace Hub path (e.g "gs://google/gemma-2-2b), or
                a local HuggingFace checkpoint path (e.g. tmp/my_model/checkpoint), or
                a GCS HuggingFace checkpoint path (e.g. gs://bucket_name/my_model/checkpoint)
            seq_len (int): Maximum sequence length.
            per_device_batch_size (int): Batch size per device.
            precision (str, optional): Precision mode for computations.
                Defaults to "mixed_bfloat16".
            scan_layers (bool, optional): Whether to use scan layers. Defaults to "auto".
                Scan layer means to stack the weights of sequential layers into one logical
                array, which will significantly speed up compilation time. However, when
                converting weights from HuggingFace into MaxText, layer scanning may led to
                OOM if the stacked weights is too large to fix in HBM.
            max_prefill_predict_length (int, optional): The maximum length of prompt tokens supported at
                inferenece time. The number of generated tokens is capped to max_seq_len -
                max_prefill_length. This number is required by the inference server. It is
                defaulted to "auto" which sets to `seq_len//2`. You should set it a larger number
                if you are running inference with this model on larger prompts; or set it to a
                smaller number if you are working with short prompts and what to generate more tokens.
            maxtext_config_args (Optional[dict], optional): Additional configuration arguments.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            MaxTextModel: A new instance of MaxTextModel initialized with pretrained weights.
        """
        set_global_model_implementation_type(ModelImplementationType.MAXTEXT)

        set_precision(precision)
        weight_dtype = cls._weight_dtype(precision)
        activation_dtype = cls._activation_dtype(precision)

        model_name = get_model_name_from_preset_handle(preset_handle)
        if scan_layers == "auto":
            scan_layers = does_model_support_scanning(model_name)
        if max_prefill_predict_length == "auto":
            max_prefill_predict_length = seq_len // 2
        maxtext_config, sharding_strategy, model = cls.initialize_random_maxtext_model(
            model_name,
            seq_len,
            per_device_batch_size,
            weight_dtype,
            activation_dtype,
            scan_layers,
            max_prefill_predict_length,
            maxtext_config_args,
        )
        model = load_hf_weights_into_maxtext_model(preset_handle, model, scan_layers)

        return cls(
            model,
            sharding_strategy,
            model_name=model_name,
            precision=precision,
            scan_layers=scan_layers,
            maxtext_config=maxtext_config,
        )

    def _convert_text_to_tokens(
        self,
        prompts: Union[str | List[str]],
        tokenizer: "AutoTokenizer",
    ) -> list[np.ndarray]:
        tokens: Dict[str, np.ndarray] = tokenizer(
            prompts,
            padding="do_not_pad",
            return_tensors="np",
        )

        return tokens["input_ids"]

    def _generate(
        self,
        inputs: Union[List[str], List[np.ndarray]],
        max_length: int = None,
        stop_token_ids: Optional[List] = None,
        strip_prompt: str = False,
        tokenizer: Optional["AutoTokenizer"] = None,
        **kwargs,
    ) -> List[List[int]]:
        """Generate tokens using the model.

        Args:
            inputs (list[str]|list[np.ndarray]): A non-empty list of strings
                or numpy arrays.
            max_length (int, optional): Maximum total sequence length
                (prompt + generated tokens).
            stop_token_ids (List[int], optional): List of token IDs that stop
                generation.
            strip_prompt (bool, optional): If True, returns only the generated
                tokens without the input prompt tokens. If False, returns all
                tokens, including the prompt tokens. Defaults to False.
            tokenizer (AutoTokenizer, optional): A HuggingFace AutoTokenizer instance.
                This is guaranteed to be provided when inputs are strings.

        Returns:
            list[list[int]]: Generated token IDs for each prompt
        """

        self._maybe_initialize_maxengine()

        if isinstance(inputs[0], str):
            inputs: list[np.ndarray] = self._convert_text_to_tokens(inputs, tokenizer)

        params = MaxTextConversionMixin.get_maxtext_params(self.model)
        if max_length is None:
            max_length = self.maxtext_config.max_target_length
        max_length = min(max_length, self.maxtext_config.max_target_length)

        def run_inference_on_one_batch(batch_prompts: list[np.ndarray]):
            # Start result buffer with the prompt tokens, later we will strip
            # prompt tokens if strip_prompt=True
            batch_results = [
                prompt.tolist() if isinstance(prompt, np.ndarray) else prompt
                for prompt in batch_prompts
            ]
            finished = [False for _ in batch_prompts]
            # Prefill with the prompt tokens
            for i, prompt_tokens in enumerate(batch_prompts):
                if len(prompt_tokens) >= min(
                    self.maxtext_config.max_prefill_predict_length, max_length
                ):
                    print(
                        f"Skipping {i}'th input because it's length "
                        "({len(prompt_tokens)}) exceeds max_length({max_length}) "
                        f"or max_prefill_predict_length({self.maxtext_config.max_prefill_predict_length})."
                    )
                    finished[i] = True
                    continue
                prefill_result, first_token = self.maxengine.prefill(
                    params=params,
                    padded_tokens=prompt_tokens,
                    true_length=len(prompt_tokens),
                )
                batch_results[i].append(int(prefill_result["tokens"][0][0]))
                self.decode_state = self.maxengine.insert(
                    prefill_result, self.decode_state, slot=i
                )

            # Generate the remaining tokens
            for _ in range(max_length):
                if all(finished):
                    break
                self.decode_state, _ = self.maxengine.generate(
                    params,
                    self.decode_state,
                )
                for i in range(len(batch_prompts)):
                    if finished[i]:
                        continue
                    token = int(self.decode_state["tokens"][i][0])
                    batch_results[i].append(token)
                    # Check if generation should continue
                    if token in stop_token_ids or len(batch_results[i]) >= max_length:
                        finished[i] = True
                if all(finished):
                    break
            return batch_results

        results = []
        total_num_prompts = len(inputs)

        for batch_start in tqdm(
            range(0, total_num_prompts, self.maxengine.max_concurrent_decodes),
            desc="Running batch inference",
            unit="batch",
        ):
            results.extend(
                run_inference_on_one_batch(
                    inputs[
                        batch_start : batch_start
                        + self.maxengine.max_concurrent_decodes
                    ]
                )
            )
        if strip_prompt:
            for i in range(total_num_prompts):
                results[i] = results[i][len(inputs[i]) :]

        return results

    def save_in_hf_format(
        self, output_dir: str, dtype: str = "auto", parallel_threads=8
    ):
        """Save the model in HuggingFace format, including the model configuration file (`config.json`),
            the model weights file (`model.safetensors` for models smaller than
            `DEFAULT_MAX_SHARD_SIZE` and `model-x-of-x.safetensors` for larger models),
            and the safe tensors index file (`model.safetensors.index.json`).

        Args:
            output_dir (str): Directory path where the model should be saved.
                Directory could be a local folder (e.g. "foldername/"), 
                HuggingFaceHub repo (e.g. "hf://your_hf_id/repo_name") or a 
                Google cloud storage path (e.g. "gs://your_bucket/folder_name), 
                and will be created if it doesn't exist. 
            dtype (str, optional): Data type for saved weights. Defaults to "auto".
            parallel_threads (int, optional): Number of parallel threads to use for saving.
                Defaults to 8. Make sure the local system has at least
                `parallel_threads * DEFAULT_MAX_SHARD_SIZE` free disk space,
                as each thread will maintain a local cache of size `DEFAULT_MAX_SHARD_SIZE`.
        """
        save_maxtext_model_in_hf_format(
            self, output_dir, dtype=dtype, parallel_threads=parallel_threads
        )

    def _maybe_initialize_maxengine(self):
        if self.maxengine is None:
            self.maxengine = MaxEngine(self.maxtext_config)
            params = MaxTextConversionMixin.get_maxtext_params(self.model)
            self.maxengine.load_existing_params(params)
            self.decode_state = self.maxengine.init_decode_state()


def does_model_support_scanning(model_name):
    return model_name in [
        supported_models.GEMMA2_2B,
        supported_models.GEMMA2_9B,
        supported_models.LLAMA31_8B,
        supported_models.LLAMA32_1B,
        supported_models.LLAMA32_3B,
    ]



#  Copyright 2025 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Script to get golden data for SFTTrainer in TRL.

Usage:

python3 MaxText/scratch_code/generate_sft_in_trl_golden_logits.py --model-name=llama3.1-8b --tokenizer-path=meta-llama/Llama-3.1-8B --max-target-length=64
"""

import argparse
import jsonlines
import os
import sys
import torch
from transformers import TrainingArguments, AutoModelForCausalLM, AutoTokenizer
from trl import SFTConfig, SFTTrainer


DATA = {
    "messages": [
        {"role": "user", "content": "Hello, what is your name?"},
        {"role": "assistant", "content": "I am a chatbot. How can I help?"},
    ],
}


def get_hf_model(tokenizer_path):
  return AutoModelForCausalLM.from_pretrained(
      tokenizer_path,
      torch_dtype=torch.float32,
  )


def get_tokenizer(tokenizer_path, max_target_length):
  return AutoTokenizer.from_pretrained(
      tokenizer_path,
      add_bos_token=False,
      add_eos_token=False,
      model_max_length=max_target_length,
  )


def setup_sft_trainer(data, hf_model, tokenizer, max_target_length):
  training_args = TrainingArguments(
      per_device_train_batch_size=1,
      bf16=True,
  )
  return SFTTrainer(
      model=hf_model,
      processing_class=tokenizer,
      train_dataset=data,
      data_collator=None,
      args=SFTConfig(
          dataset_kwargs={"skip_prepare_dataset": True},
          max_seq_length=max_target_length,
          **training_args.to_dict(),
      ),
  )


def apply_chat_template():
  messages = []
  for message in DATA["messages"]:
    if message["role"] == "user":
      messages.append("<user>" + message["content"] + "</user>")
    elif message["role"] == "assistant":
      messages.append("<assistant>" + message["content"] + "</assistant>")
  return messages


def get_input_ids(data, tokenizer, max_target_length):
  input_ids = []
  attention_mask = []
  for d in data:
    input_ids += d["input_ids"]
    attention_mask += d["attention_mask"]
  labels = input_ids + [tokenizer.eos_token_id] + [0]
  input_ids = [tokenizer.bos_token_id] + input_ids + [tokenizer.eos_token_id]
  attention_mask = [1] + attention_mask + [1]
  return {
      "input_ids": torch.tensor(input_ids[:max_target_length], dtype=torch.long).unsqueeze(0),
      "labels": torch.tensor(labels[:max_target_length], dtype=torch.long).unsqueeze(0),
      "attention_mask": torch.tensor(attention_mask[:max_target_length], dtype=torch.long).unsqueeze(0),
  }


def prepare_trl_inputs(tokenizer, max_target_length):
  data = apply_chat_template()
  tokenized_data = [tokenizer(d) for d in data]
  processed_data = get_input_ids(tokenized_data, tokenizer, max_target_length)
  return processed_data


def save_golden_logits(config):
  hf_model = get_hf_model(config.tokenizer_path)
  tokenizer = get_tokenizer(config.tokenizer_path, config.max_target_length)
  trl_data = prepare_trl_inputs(tokenizer, config.max_target_length)
  trl_trainer = setup_sft_trainer(trl_data, hf_model, tokenizer, config.max_target_length)
  _, trl_outputs = trl_trainer.compute_loss(hf_model, trl_data, return_outputs=True)
  trl_logits = trl_outputs.logits.detach().numpy()

  data_to_save = {
      "data": DATA,
      "tokens": trl_data["input_ids"][0].tolist(),
      "attention_mask": trl_data["attention_mask"][0].tolist(),
      "trl_logits": trl_logits.tolist()[0],
  }

  model_output_path = "/".join([os.getcwd(), "MaxText/test_assets", f"golden_data_sft_{config.model_name}.jsonl"])
  with jsonlines.open(model_output_path, "w") as f:
    f.write(data_to_save)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--model-name", type=str, required=False, default="llama3.1-8b")
  parser.add_argument("--tokenizer-path", type=str, required=False, default="meta-llama/Llama-3.1-8B")
  parser.add_argument("--max-target-length", type=int, required=False, default=64)
  config = parser.parse_args(sys.argv[1:])
  save_golden_logits(config)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This software may be used and distributed in accordance with the terms of the
# Llama 3 Community License Agreement.

"""Tiktoken tokenizer used for llama3."""

import os
from logging import getLogger
from pathlib import Path
from typing import (
    AbstractSet,
    cast,
    Collection,
    Dict,
    Iterator,
    List,
    Literal,
    Sequence,
    TypedDict,
    Union,
)

import tiktoken
from tiktoken.load import load_tiktoken_bpe


logger = getLogger(__name__)


Role = Literal["system", "user", "assistant"]


class Message(TypedDict):
  role: Role
  content: str


Dialog = Sequence[Message]


class Tokenizer:
  """
  Tokenizing and encoding/decoding text using the Tiktoken tokenizer.
  """

  special_tokens: Dict[str, int]

  num_reserved_special_tokens = 256

  # pylint: disable=line-too-long
  pat_str = r"(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}{1,3}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"  # noqa: E501

  def __init__(self, model_path: str):
    """
    Initializes the Tokenizer with a Tiktoken model.
    Args:
        model_path (str): The path to the Tiktoken model file.
    """
    assert os.path.isfile(model_path), model_path

    mergeable_ranks = load_tiktoken_bpe(model_path)
    num_base_tokens = len(mergeable_ranks)
    special_tokens = [
        "<|begin_of_text|>",
        "<|end_of_text|>",
        "<|reserved_special_token_0|>",
        "<|reserved_special_token_1|>",
        "<|reserved_special_token_2|>",
        "<|reserved_special_token_3|>",
        "<|start_header_id|>",
        "<|end_header_id|>",
        "<|reserved_special_token_4|>",
        "<|eot_id|>",  # end of turn
    ] + [
        f"<|reserved_special_token_{i}|>"
        for i in range(5, self.num_reserved_special_tokens - 5)
    ]
    self.special_tokens = {
        token: num_base_tokens + i for i, token in enumerate(special_tokens)
    }
    self.model = tiktoken.Encoding(
        name=Path(model_path).name,
        pat_str=self.pat_str,
        mergeable_ranks=mergeable_ranks,
        special_tokens=self.special_tokens,
    )
    logger.info("Reloaded tiktoken model from %s", model_path)

    self.n_words: int = self.model.n_vocab
    # BOS / EOS token IDs
    self.bos_id: int = self.special_tokens["<|begin_of_text|>"]
    self.eos_id: int = self.special_tokens["<|end_of_text|>"]
    self.pad_id: int = -1
    self.stop_tokens = {
        self.special_tokens["<|end_of_text|>"],
        self.special_tokens["<|eot_id|>"],
    }

    logger.info(
        "#words: %d - BOS ID: %d - EOS ID: %d",
        self.n_words,
        self.bos_id,
        self.eos_id,
    )

  def encode(
      self,
      s: str,
      *,
      bos: bool = False,
      eos: bool = False,
      allowed_special: Union[Literal["all"], AbstractSet[str]] | None = None,
      disallowed_special: Union[Literal["all"], Collection[str]] = (),
  ) -> List[int]:
    """
    Encodes a string into a list of token IDs.
    Args:
        s (str): The input string to be encoded.
        bos (bool): Whether to prepend the beginning-of-sequence token.
        eos (bool): Whether to append the end-of-sequence token.
        allowed_tokens ("all"|set[str]): allowed special tokens in string
        disallowed_tokens ("all"|set[str]): special tokens that raise an error when in string
    Returns:
        list[int]: A list of token IDs.
    By default, setting disallowed_special=() encodes a string by ignoring
    special tokens. Specifically:
    - Setting `disallowed_special` to () will cause all text corresponding
      to special tokens to be encoded as natural text (insteading of raising
      an error).
    - Setting `allowed_special` to "all" will treat all text corresponding
      to special tokens to be encoded as special tokens.
    """
    assert isinstance(s, str)

    # The tiktoken tokenizer can handle <=400k chars without
    # pyo3_runtime.PanicException.
    tiktoken_max_encode_chars = 400_000

    # https://github.com/openai/tiktoken/issues/195
    # Here we iterate over subsequences and split if we exceed the limit
    # of max consecutive non-whitespace or whitespace characters.
    max_no_whitespaces_chars = 25_000

    substrs = (
        substr
        for i in range(0, len(s), tiktoken_max_encode_chars)
        for substr in self._split_whitespaces_or_nonwhitespaces(
            s[i : i + tiktoken_max_encode_chars], max_no_whitespaces_chars
        )
    )
    t: List[int] = []
    if allowed_special is None:
      allowed_special = set()
    for substr in substrs:
      t.extend(
          self.model.encode(
              substr,
              allowed_special=allowed_special,
              disallowed_special=disallowed_special,
          )
      )
    if bos:
      t.insert(0, self.bos_id)
    if eos:
      t.append(self.eos_id)
    return t

  def decode(self, t: Sequence[int]) -> str:
    """
    Decodes a list of token IDs into a string.
    Args:
        t (List[int]): The list of token IDs to be decoded.
    Returns:
        str: The decoded string.
    """
    # Typecast is safe here. Tiktoken doesn't do anything list-related with the sequence.
    return self.model.decode(cast(List[int], t))

  @staticmethod
  def _split_whitespaces_or_nonwhitespaces(
      s: str, max_consecutive_slice_len: int
  ) -> Iterator[str]:
    """
    Splits the string `s` so that each substring contains no more than `max_consecutive_slice_len`
    consecutive whitespaces or consecutive non-whitespaces.
    """
    current_slice_len = 0
    current_slice_is_space = s[0].isspace() if len(s) > 0 else False
    slice_start = 0

    for i in range(len(s)):
      is_now_space = s[i].isspace()

      if current_slice_is_space ^ is_now_space:
        current_slice_len = 1
        current_slice_is_space = is_now_space
      else:
        current_slice_len += 1
        if current_slice_len > max_consecutive_slice_len:
          yield s[slice_start:i]
          slice_start = i
          current_slice_len = 1
    yield s[slice_start:]


class ChatFormat:
  """Helper class to encode/decode messages in chat format."""

  def __init__(self, tokenizer: Tokenizer):
    self.tokenizer = tokenizer

  def encode_header(self, message: Message) -> List[int]:
    tokens = []
    tokens.append(self.tokenizer.special_tokens["<|start_header_id|>"])
    tokens.extend(self.tokenizer.encode(message["role"], bos=False, eos=False))
    tokens.append(self.tokenizer.special_tokens["<|end_header_id|>"])
    tokens.extend(self.tokenizer.encode("\n\n", bos=False, eos=False))
    return tokens

  def encode_message(self, message: Message) -> List[int]:
    tokens = self.encode_header(message)
    tokens.extend(
        self.tokenizer.encode(message["content"].strip(), bos=False, eos=False)
    )
    tokens.append(self.tokenizer.special_tokens["<|eot_id|>"])
    return tokens

  def encode_dialog_prompt(self, dialog: Dialog) -> List[int]:
    tokens = []
    tokens.append(self.tokenizer.special_tokens["<|begin_of_text|>"])
    for message in dialog:
      tokens.extend(self.encode_message(message))
    # Add the start of an assistant message for the model to complete.
    tokens.extend(self.encode_header({"role": "assistant", "content": ""}))
    return tokens



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import enum
import dataclasses
import jax


# TODO: change to enum.StrEnum after Python upgrading in Google Cloud TPU.
@enum.unique
class ParallelAxis(enum.Enum):
  # X is used for Tensor Parallelism (as major axis) and Expert Parallelism.
  X = enum.auto()
  # Y is used for Sequence, Sequence Pipeline, Decode Batch and Tensor Parallelism (as minor axis).
  Y = enum.auto()


def dp_axis_names():
  return (ParallelAxis.Y.name,)


def sp_axis_names():
  return (ParallelAxis.Y.name,)


def spp_axis_names():
  return (ParallelAxis.Y.name,)


def tp_axis_names():
  return (ParallelAxis.X.name, ParallelAxis.Y.name)


def tp_major_axis_names():
  return (ParallelAxis.X.name,)


def tp_minor_axis_names():
  return (ParallelAxis.Y.name,)


@enum.unique
class ModelParallelStrategy(enum.Enum):
  """Overall Transformer Parallel Strategy."""

  TENSOR_PARALLEL = enum.auto()


@dataclasses.dataclass
class ModelParallelConfig:
  mesh: jax.sharding.Mesh
  parallel_type: ModelParallelStrategy = ModelParallelStrategy.TENSOR_PARALLEL


@enum.unique
class FFWParallelStrategy(enum.Enum):
  """Overall Transformer FFW Layer Parallel Strategy.

  Please refer to https://arxiv.org/pdf/2211.05102"""

  ONE_D_WEIGHT_STATIONARY = enum.auto()


@dataclasses.dataclass
class FeedForwardParallelConfig:
  mesh: jax.sharding.Mesh
  parallel_type: FFWParallelStrategy | None = None
  enable_collective_matmul: bool = False


@dataclasses.dataclass
class AttentionParallelConfig:
  mesh: jax.sharding.Mesh
  gather_input: bool = False
  reduce_output: bool = False


@enum.unique
class LinearParallelType(enum.Enum):
  """Parallel Type for Linear Layer weight."""

  ROW = enum.auto()
  COLUMN = enum.auto()


@enum.unique
class CollectiveMatmulType(enum.Enum):
  ALL_GATHER = enum.auto()
  REDUCE_SCATTER = enum.auto()


@dataclasses.dataclass
class LinearParallelConfig:
  mesh: jax.sharding.Mesh
  parallel_type: LinearParallelType | None = None
  reduce_scatter_output: bool = False
  reduce_output: bool = False
  collective_matmul_type: CollectiveMatmulType | None = None


@dataclasses.dataclass
class RMSNormParallelConfig:
  mesh: jax.sharding.Mesh
  activation_sharded: bool = False


@enum.unique
class EmbeddingParallelType(enum.Enum):
  """Parallel Type for Embedding Layer weight."""

  COLUMN = enum.auto()


@dataclasses.dataclass
class EmbeddingParallelConfig:
  mesh: jax.sharding.Mesh
  parallel_type: EmbeddingParallelType | None = None


@dataclasses.dataclass
class DecoderLayerParallelConfig:
  mesh: jax.sharding.Mesh



"""
Copyright 2024 Google LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

"""offline driver module"""


import datetime
import multiprocessing as mp
import threading
import queue
import math
from typing import Sequence
import jax
from inference import parallel
from inference.config.config import Config
from inference.runtime.engine import Engine, EngineMode, OfflineChannel
from inference.runtime.request_type import *


class OfflineInference:

  def __init__(
      self,
      model_id: str,
      num_engines: int,
      enable_multiprocessing: bool,
  ):
    self.num_engines = num_engines
    self.req_queues: list[mp.Queue | queue.Queue] = []
    self.res_queues: list[mp.Queue | queue.Queue] = []
    self._running_pool: list[mp.Process | threading.Thread] = []
    self._engine_started_events: list = []
    self._engine_completed_events: list = []
    for i in range(num_engines):
      # Create queues, events for a Process/Thread runner
      if enable_multiprocessing:
        req_q, res_q = mp.Queue(), mp.Queue()
        started, completed = mp.Event(), mp.Event()
        runner = mp.Process
      else:
        req_q, res_q = queue.Queue(), queue.Queue()
        started, completed = threading.Event(), threading.Event()
        runner = threading.Thread
      # Config the runner
      self.req_queues.append(req_q)
      self.res_queues.append(res_q)
      self._engine_started_events.append(started)
      self._engine_completed_events.append(completed)
      execution = runner(
          target=self.launch_engine,
          args=(
              model_id,
              self.req_queues[i],
              self.res_queues[i],
              self._engine_started_events[i],
              self._engine_completed_events[i],
          ),
      )
      self._running_pool.append(execution)

    # Launch the runners
    for e in self._running_pool:
      e.start()

    for started_event in self._engine_started_events:
      while not started_event.is_set():
        started_event.wait()

  def launch_engine(
      self,
      model_id,
      req_queue,
      res_queue,
      started_event,
      completion_event,
  ):
    devices = jax.devices()
    mesh = parallel.create_device_mesh(
        devices=devices,
        shape=(len(devices), 1),
    )
    engine = Engine(
        mesh=mesh,
        inference_params=Config.get(model_id),
        mode=EngineMode.OFFLINE,
        channel=OfflineChannel(
            req_queue=req_queue,
            res_queue=res_queue,
        ),
        debug_mode=False,
    )
    engine.start()
    started_event.set()
    while not completion_event.is_set():
      completion_event.wait()
    engine.stop()
    return

  def __call__(self, prompts: Sequence[str]) -> tuple[list[Response], float]:
    for i in range(self.num_engines):
      while not self._engine_started_events[i].is_set():
        self._engine_started_events[i].wait()

    print(f"Offline inference begins: {datetime.datetime.now()} ...")

    num_prompts = len(prompts)
    for p in range(num_prompts):
      i = p % self.num_engines
      self.req_queues[i].put(OfflineRequest(prompt=prompts[p]))

    res = []
    for p in range(num_prompts):
      i = p % self.num_engines
      res.append(self.res_queues[i].get())

    print("Offline inference ends:", datetime.datetime.now())

    for i in range(self.num_engines):
      self._engine_completed_events[i].set()

    for e in self._running_pool:
      e.join()

    return res
